"""
See https://learn.microsoft.com/en-us/entra/identity-platform/access-tokens
"""

import logging
import re

import httpx
from aiocache import cached
from fastapi import Depends, Header, status
from fastapi.exceptions import HTTPException
from joserfc import jwt
from joserfc.jwk import KeySet
from joserfc.jwt import JWTClaimsRegistry

from ... import models
from ...config import settings

# Try to import jwks from project directory first (user override)
# Fall back to package location (default implementation)
try:
    from auth.entraid import jwks
except ModuleNotFoundError:
    from . import jwks

logger = logging.getLogger(__name__)

OPENID_CONNECT_DISCOVERY_DOCUMENT_URL = (
    "https://login.microsoftonline.com/common/v2.0/.well-known/openid-configuration"
)


@cached(ttl=60 * 60 * 24)
async def get_jwks_json_default():
    logger.info("Get default JWKS json for Entra ID")
    async with httpx.AsyncClient() as client:
        response = await client.get(OPENID_CONNECT_DISCOVERY_DOCUMENT_URL)
    jwks_uri = response.json()["jwks_uri"]
    async with httpx.AsyncClient() as client:
        response = await client.get(jwks_uri)
    return response.json()


async def get_key_set():
    jwks_data = await jwks.get_jwks_json()
    if jwks_data is None:
        jwks_data = await get_jwks_json_default()
    key_set = KeySet.import_key_set(jwks_data)
    return key_set


@cached(ttl=60 * 60)
async def validate_token(token_string: str):
    """Validates the Entra ID access/id token. Returns a user object."""
    logger.debug(f"Validating token: {token_string}")
    if token_string.lower().startswith("error"):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail=f"Auth error with Entra ID token: {token_string} See https://learn.microsoft.com/en-us/office/dev/add-ins/develop/troubleshoot-sso-in-office-add-ins#causes-and-handling-of-errors-from-getaccesstoken",
        )
    if token_string.lower().startswith("bearer"):
        parts = token_string.split()
        if len(parts) != 2:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Auth error: Invalid token, must be in format 'Bearer xxxx'",
            )
        token_string = parts[1]
    else:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Auth error: Invalid token, must be in format 'Bearer xxxx'",
        )
    key_set = await get_key_set()
    try:
        token = jwt.decode(token_string, key_set)
    except Exception:
        logger.exception("Auth error: Failed to decode token")
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Auth error: Failed to decode token",
        )
    token_version = token.claims.get("ver")
    # https://learn.microsoft.com/en-us/azure/active-directory/develop/access-tokens#token-formats
    # Upgrade to 2.0:
    # https://learn.microsoft.com/en-us/answers/questions/639834/how-to-get-access-token-version-20.html
    if token_version == "1.0":
        issuer = f"https://sts.windows.net/{settings.auth_entraid_tenant_id}/"
        audience = f"api://{settings.auth_entraid_client_id}"
    elif token_version == "2.0":
        issuer = (
            f"https://login.microsoftonline.com/{settings.auth_entraid_tenant_id}/v2.0"
        )
        audience = settings.auth_entraid_client_id
    else:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail=f"Auth error: Unsupported token version: {token_version}",
        )
    try:
        if not settings.auth_entraid_multitenant:
            claims_requests = JWTClaimsRegistry(
                aud={"essential": True, "value": audience},
                iss={"essential": True, "value": issuer},
            )
            claims_requests.validate(token.claims)
        else:
            claims_requests = JWTClaimsRegistry(
                aud={"essential": True, "value": audience},
            )
            claims_requests.validate(token.claims)
            # External users have their own tenant_id
            issuer_regex = r"https://login\.microsoftonline\.com/(common|organizations|consumers|[0-9a-fA-F-]{36})/v2\.0"
            if not re.match(issuer_regex, issuer):
                logger.debug(f"Couldn't match issuer for token: {token_string}")
                raise HTTPException(
                    status_code=status.HTTP_401_UNAUTHORIZED,
                    detail="Auth error: Couldn't validate token",
                )
        logger.debug(claims_requests)
    except Exception as e:
        logger.debug(f"Authentication error for token: {token_string}")
        logger.info(repr(e))
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Auth error: Couldn't validate token",
        )

    current_user = models.User(claims=token.claims)
    logger.info(f"User authenticated: {current_user.name}")
    return current_user



import importlib
import json
import logging
from typing import Annotated, Union

import redis
import xlwings as xw
from fastapi import Depends, Form, Header, HTTPException, Request, status

from . import models
from .config import settings
from .databases import get_redis_client

logger = logging.getLogger(__name__)


# Book
async def parse_book_input(
    request: Request,
    form_data: str | None = Form(None, alias="bookData"),
) -> dict:
    """Helper dependency to parse either form data (htmx)
    or body (custom scripts & custom functions) -- couldn't make Body() work"""
    if form_data:
        return json.loads(form_data)
    else:
        body_bytes = await request.body()
        if body_bytes:
            body_str = body_bytes.decode("utf-8")
            return json.loads(body_str)
    raise HTTPException(status_code=400, detail="No book data provided")


async def get_book(book_data: dict = Depends(parse_book_input)):
    """Book dependency that returns the calling book and cleans it up again"""
    book = xw.Book(json=book_data)
    try:
        yield book
    finally:
        book.close()


Book = Annotated[xw.Book, Depends(get_book)]


# Users/Auth
async def authenticate(
    token_string: str = Header(default="", alias="Authorization"),
    auth_provider: Union[str, None] = Header(default=None),
):
    if not settings.auth_providers:
        return User(id="n/a", name="Anonymous")
    if len(settings.auth_providers) == 1:
        provider = settings.auth_providers[0]
    elif len(settings.auth_providers) > 1 and not auth_provider:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="With multiple auth providers, you need to provide the Auth-Provider header.",
        )
    elif auth_provider not in settings.auth_providers:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="Auth-Provider header wasn't found in XLWINGS_AUTH_PROVIDERS setting.",
        )
    else:
        provider = auth_provider
    logger.info(f"Using auth provider {provider}")
    # Validate the provider before import to prevent non-literal-import SAST flagging
    if provider not in settings.auth_providers:
        raise ValueError(f"Unsupported authentication provider: {provider}")
    try:
        # Try to import from project directory first (user override)
        try:
            module = importlib.import_module(f"auth.{provider}")
        except ModuleNotFoundError:
            # Fall back to package location (default implementation)
            module = importlib.import_module(f"xlwings_server.auth.{provider}")
        current_user = await module.validate_token(token_string)
    except (AttributeError, ModuleNotFoundError):
        logger.exception(f"Auth provider '{provider}' implementation missing.")
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f"Auth provider '{provider}' implementation missing.",
        )
    return current_user


async def get_user(request: Request, current_user: models.User = Depends(authenticate)):
    # Extract IP address and attach it to the user object
    if "x-forwarded-for" in request.headers:
        ip_address = request.headers["x-forwarded-for"].split(",")[0].strip()
    else:
        ip_address = request.client.host if request.client else None
    current_user.ip_address = ip_address

    if not settings.auth_providers:
        return current_user
    # RBAC
    has_required_roles = await current_user.has_required_roles(
        settings.auth_required_roles
    )
    if not has_required_roles:
        missing_roles = ", ".join(
            set(settings.auth_required_roles).difference(current_user.roles)
        )
        msg = f"Auth error: Missing roles for {current_user.name}: {missing_roles}"
        logger.warning(msg)
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail=msg,
        )
    # Authorization
    if not await current_user.is_authorized():
        msg = f"Auth error: Not authorized for {current_user.name}"
        logger.warning(msg)
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail=msg,
        )
    return current_user


User = Annotated[models.User, Depends(get_user)]

# Redis
RedisClient = Annotated[redis.Redis, Depends(get_redis_client)]



"""
Uvicorn reloads the app when the Python code is changed. So, we need to reload the
browser when the app starts to show any changes caused the Python code.
But, restarting the app takes a lot of time (especially in Docker), so we
watch HTML, CSS, and JS files separately via Watchfiles. If these files change, we just
reload the browser without having to restart the Python app.
"""

from pathlib import Path

from watchfiles import Change, DefaultFilter, awatch

from .config import settings

browser_reload_triggered_by_backend = False
watching_frontend_files = False


class WebFilter(DefaultFilter):
    allowed_extensions = (".html", ".css", ".js", ".py", ".txt", ".env")

    def __call__(self, change: Change, path: str) -> bool:
        if not super().__call__(change, path):
            return False

        path = Path(path)
        if path.suffix not in self.allowed_extensions:
            return False

        # Allow HTML/CSS/JS files anywhere
        if path.suffix in (".html", ".css", ".js"):
            return True

        # xlwings Wasm
        if settings.enable_wasm and path.suffix in (".py", ".txt", ".env"):
            allowed_dirs = ("wasm", "custom_scripts", "custom_functions")
            return any(dir_name in path.parts for dir_name in allowed_dirs)

        return False


async def watch_frontend_files(sio, directory):
    async for changes in awatch(
        directory,
        watch_filter=WebFilter(),
    ):
        await sio.emit("reload")


async def start_browser_reload_watcher(sio, directory):
    """Needs to be called from the sio connect event on the backend"""
    global browser_reload_triggered_by_backend
    global watching_frontend_files
    if not browser_reload_triggered_by_backend and not settings.enable_wasm:
        await sio.emit("reload")
        browser_reload_triggered_by_backend = True
    if not watching_frontend_files:
        sio.start_background_task(watch_frontend_files, sio=sio, directory=directory)
        watching_frontend_files = True



import logging
import zlib

try:
    import numpy as np
except ImportError:
    np = None
try:
    import pandas as pd
except ImportError:
    pd = None
import redis
from croniter import croniter
from xlwings import XlwingsError
from xlwings.constants import ObjectHandleIcons
from xlwings.conversion import Converter

from .config import settings
from .routers import xlwings as xlwings_router
from .serializers import deserialize, serialize

logger = logging.getLogger(__name__)

# Used if XLWINGS_OBJECT_CACHE_URL, i.e., Redis isn't configured.
# Only useful with a single worker e.g., during development.
cache = {}


class ObjectCacheConverter(Converter):
    @staticmethod
    def read_value(cell_address, options):
        # For custom function args of type Entity, the frontend sends the cell address
        # instead of the cell value
        redis_client: redis.Redis = xlwings_router.redis_client_context.get()
        if settings.object_cache_url and not redis_client:
            raise XlwingsError("Failed to connect to Redis")
        key = f"object:{cell_address}"
        if settings.object_cache_url:
            value = redis_client.get(key)
            if not value:
                raise XlwingsError("Object cache is empty")
            if settings.object_cache_enable_compression:
                value = zlib.decompress(value).decode()
            else:
                value = value.decode()
        else:
            value = cache.get(key)
            if not value:
                raise XlwingsError("Object cache is empty")
        obj = deserialize(value)
        return obj

    @staticmethod
    def write_value(obj, options):
        redis_client: redis.Redis = xlwings_router.redis_client_context.get()
        if settings.object_cache_url and not redis_client:
            raise XlwingsError("Failed to connect to Redis")
        key = f"object:{xlwings_router.caller_address_context.get()}"
        values = serialize(obj)
        if settings.object_cache_url:
            expire_at = None
            if settings.object_cache_expire_at:
                cron = croniter(settings.object_cache_expire_at)
                expire_at = int(cron.get_next())
            if settings.object_cache_enable_compression:
                values = zlib.compress(values.encode())
            redis_client.set(key, values, exat=expire_at)
        else:
            logger.warning(
                "Storing objects in memory. Configure XLWINGS_OBJECT_CACHE_URL "
                "for production use!"
            )
            cache[key] = values

        obj_type = type(obj).__name__

        result = {
            "type": "Entity",
            "text": options.get("text", obj_type) or obj_type,
            "properties": {
                "Type": {
                    "type": "String",
                    "basicValue": obj_type,
                },
            },
            "layouts": {
                "compact": {
                    "icon": options.get("icon", ObjectHandleIcons.generic)
                    or ObjectHandleIcons.generic
                }
            },
        }

        # Shape
        def get_shape(obj):
            if pd and isinstance(obj, pd.DataFrame):
                return f"{obj.shape}"
            if np and isinstance(obj, np.ndarray):
                return f"{obj.shape}"
            elif isinstance(obj, (list, tuple)):
                if obj and isinstance(obj[0], (list, tuple)):
                    return f"({len(obj)}, {len(obj[0])})"
                return f"({len(obj)},)"
            else:
                try:
                    return f"{len(obj)} (length)"
                except Exception:
                    return None

        shape_value = get_shape(obj)
        if shape_value:
            result["properties"]["Shape"] = {
                "type": "String",
                "basicValue": shape_value,
            }

        # Columns
        cols_info = None
        if pd and isinstance(obj, pd.DataFrame):
            cols_info = ", ".join(f"{col} [{obj[col].dtype}]" for col in obj.columns)
        if cols_info:
            result["properties"]["Columns"] = {
                "type": "String",
                "basicValue": cols_info,
            }

        # Index
        index_info = None
        if pd and isinstance(obj, pd.DataFrame):
            index_type = type(obj.index).__name__
            index_length = len(obj.index)
            index_start = obj.index[0]
            index_end = obj.index[-1]
            index_info = (
                f"{index_type}: {index_length} entries, {index_start} to {index_end}"
            )
        if index_info:
            result["properties"]["Index"] = {
                "type": "String",
                "basicValue": index_info,
            }
        return result



import contextvars
import inspect
import logging
from pathlib import Path
from textwrap import dedent

import xlwings as xw
import xlwings.server
from fastapi import APIRouter, Body, Header, Request, Response

# Try to import custom modules from project directory first (CLI/Azure mode)
# Fall back to package location (tests/package mode)
try:
    import custom_functions
except ModuleNotFoundError:
    import xlwings_server.custom_functions as custom_functions

try:
    import custom_scripts
except ModuleNotFoundError:
    import xlwings_server.custom_scripts as custom_scripts

from xlwings_server import dependencies as dep
from xlwings_server.config import PACKAGE_DIR, settings
from xlwings_server.models import CurrentUser
from xlwings_server.templates import TemplateResponse

logger = logging.getLogger(__name__)

router = APIRouter(prefix=f"{settings.app_path}/xlwings")


def sanitize_log_input(input_string: str) -> str:
    """Replaces newline and carriage return characters to prevent log injection."""
    if not isinstance(input_string, str):
        input_string = str(input_string)
    return input_string.replace("\n", "\\n").replace("\r", "\\r")


@router.get("/alert")
async def alert(
    request: Request, prompt: str, title: str, buttons: str, mode: str, callback: str
):
    """Boilerplate required by book.app.alert() and to show unhandled exceptions"""
    return TemplateResponse(
        request=request,
        name="xlwings_alert.html",
        context={
            "prompt": prompt,
            "title": title,
            "buttons": buttons,
            "mode": mode,
            "callback": callback,
            "settings": settings,
        },
    )


@router.get("/custom-functions-meta")
@router.get("/custom-functions-meta.json")
async def custom_functions_meta():
    return xlwings.server.custom_functions_meta(
        custom_functions, typehinted_params_to_exclude=[CurrentUser]
    )


@router.get("/custom-functions-code")
@router.get("/custom-functions-code.js")
async def custom_functions_code():
    custom_functions_call_path = f"{settings.app_path}/xlwings/custom-functions-call"
    js = (
        PACKAGE_DIR / "static" / "js" / "core" / "custom-functions-code.js"
    ).read_text()
    # format string would require to double all curly braces
    js = js.replace("placeholder_xlwings_version", xw.__version__).replace(
        "placeholder_custom_functions_call_path", custom_functions_call_path
    )
    for name, obj in inspect.getmembers(custom_functions):
        if hasattr(obj, "__xlfunc__"):
            xlfunc = obj.__xlfunc__
            func_name = xlfunc["name"]
            streaming = "true" if inspect.isasyncgenfunction(obj) else "false"
            js += dedent(
                f"""\
            async function {func_name}() {{
                let args = ["{func_name}", {streaming}]
                args.push.apply(args, arguments);
                return await base.apply(null, args);
            }}
            CustomFunctions.associate("{func_name.upper()}", {func_name});
            """
            )
    return Response(
        content=js,
        media_type="text/javascript",
    )


# ContextVars
socketio_id_context = contextvars.ContextVar("socketio_id_context")
caller_address_context = contextvars.ContextVar("caller_address_context")
redis_client_context = contextvars.ContextVar("redis_client_context")


@router.post("/custom-functions-call")
async def custom_functions_call(
    current_user: dep.User,
    redis_client: dep.RedisClient,
    data: dict = Body,
    sid: str | None = Header(default=None),
):
    # Replace newline and carriage return characters to prevent log injection
    safe_func_name = sanitize_log_input(data["func_name"])
    safe_user_name = sanitize_log_input(current_user.name)
    logger.info(f"""Function "{safe_func_name}" called by {safe_user_name}""")
    socketio_id_context.set(sid)  # For utils.trigger_script()
    caller_address_context.set(data["caller_address"])  # For ObjectCache converter
    redis_client_context.set(redis_client)  # For ObjectCache converter

    rv = await xlwings.server.custom_functions_call(
        data,
        custom_functions,
        current_user,
        typehint_to_value={CurrentUser: current_user},
    )
    return {"result": rv}


@router.post("/custom-scripts-call/{script_name}")
async def custom_scripts_call(script_name: str, book: dep.Book, current_user: dep.User):
    # Replace newline and carriage return characters to prevent log injection
    safe_script_name = sanitize_log_input(script_name)
    safe_user_name = sanitize_log_input(current_user.name)
    logger.info(f"""Script "{safe_script_name}" called by {safe_user_name}""")
    book = await xlwings.server.custom_scripts_call(
        custom_scripts,
        script_name,
        current_user,
        typehint_to_value={CurrentUser: current_user, xw.Book: book},
    )
    return book.json()


@router.get("/custom-scripts-meta")
@router.get("/custom-scripts-meta.json")
async def custom_scripts_meta():
    return xlwings.server.custom_scripts_meta(custom_scripts)


if settings.enable_wasm:

    @router.get("/pyodide.json")
    async def get_pyodide_config():
        # requirements.txt
        packages = (
            Path(settings.project_dir / "wasm" / "requirements.txt")
            .read_text()
            .splitlines()
        )
        packages = [
            pkg.replace("/static", settings.static_url_path).strip()
            for pkg in packages
            if pkg.strip()
        ]

        # Files
        def scan_directory(
            base_dir: Path, dir_name: str, prepend_dir_name: bool = False
        ) -> dict:
            dir_path = Path(settings.project_dir / dir_name)
            files = {}
            if dir_path.exists():
                for file_path in dir_path.rglob("*"):
                    if (
                        file_path.is_file()
                        and file_path.suffix != ".pyc"
                        and file_path.name != "requirements.txt"
                    ):
                        relative_path = file_path.relative_to(dir_path)
                        files[
                            f"{settings.static_url_path.replace('static', dir_name)}/{relative_path}"
                        ] = (
                            f"./{dir_name}/{relative_path}"
                            if prepend_dir_name
                            else f"./{relative_path}"
                        )
            return files

        # Scan all directories
        files = {}
        files.update(scan_directory(settings.project_dir, "wasm"))
        for directory in ["custom_functions", "custom_scripts"]:
            files.update(
                scan_directory(settings.project_dir, directory, prepend_dir_name=True)
            )
        response = {"packages": packages, "files": files}
        return response



import logging
import os
import sys
import warnings
from importlib.metadata import version as get_package_version
from pathlib import Path
from typing import Any, Literal

from pydantic import UUID4, computed_field
from pydantic_settings import (
    BaseSettings,
    PydanticBaseSettingsSource,
    PyprojectTomlConfigSettingsSource,
    SettingsConfigDict,
)

logger = logging.getLogger(__name__)

# Set before xlwings is imported elsewhere in the application
os.environ["XLWINGS_ON_SERVER"] = "true"

# Get project directory from environment (set by CLI)
# Falls back to current working directory if not set (for backward compatibility)
PROJECT_DIR = Path(os.getenv("XLWINGS_PROJECT_DIR", Path.cwd()))

# Get package directory - must be calculated before any sys.path manipulation
# Use __file__ which points to the actual config.py location (in site-packages after install)
PACKAGE_DIR = Path(__file__).parent.resolve()

# Setup sys.path for user overrides
# This is also done in main.py, but we need it here too because config.py
# can be imported directly (e.g., from custom_functions) before main.py runs
if str(PROJECT_DIR) not in sys.path:
    sys.path.insert(0, str(PROJECT_DIR))

# Settings prefix used for environment variables (required)
# and pyproject.toml keys (optional)
ENV_PREFIX = "XLWINGS_"


class CaseInsensitivePyprojectTomlSource(PyprojectTomlConfigSettingsSource):
    """
    PyprojectToml source that handles case-insensitive keys and strips ENV_PREFIX.
    E.g., allows 'environment', 'ENVIRONMENT', 'XLWINGS_ENVIRONMENT', etc. to all work.
    """

    def __call__(self) -> dict[str, Any]:
        config = super().__call__()
        normalized = {}
        prefix_lower = ENV_PREFIX.lower()
        for k, v in config.items():
            # Convert to lowercase
            key = k.lower()
            # Strip prefix if present (e.g., "xlwings_")
            if key.startswith(prefix_lower):
                key = key[len(prefix_lower) :]
            normalized[key] = v
        return normalized


class Settings(BaseSettings):
    """
    xlwings Server settings with support for multiple configuration sources.

    Configuration priority (highest to lowest):
    1. Environment variables (XLWINGS_*)
    2. .env file
    3. pyproject.toml [tool.xlwings_server] section
    4. Default values defined in this class

    See .env.template for detailed documentation of all settings.
    """

    def __init__(self, **values):
        super().__init__(**values)
        if self.public_addin_store is not None:
            warnings.warn(
                "The 'XLWINGS_PUBLIC_ADDIN_STORE' field is deprecated and will be removed in "
                "future versions. Use 'XLWINGS_CDN_OFFICEJS' instead.",
                DeprecationWarning,
            )
            self.cdn_officejs = self.public_addin_store

    model_config = SettingsConfigDict(
        env_prefix=ENV_PREFIX,
        env_file=os.getenv("DOTENV_PATH", PROJECT_DIR / ".env"),
        extra="ignore",
        pyproject_toml_table_header=("tool", "xlwings_server"),
    )

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        """
        Define the sources and their priority order.
        Priority (highest to lowest):
        1. init_settings - arguments passed to Settings()
        2. env_settings - environment variables
        3. dotenv_settings - .env file
        4. pyproject.toml - [tool.xlwings_server] section
        5. file_secret_settings - secret files
        """
        # CaseInsensitivePyprojectTomlSource looks in PROJECT_DIR for pyproject.toml
        pyproject_path = PROJECT_DIR / "pyproject.toml"
        return (
            init_settings,
            env_settings,
            dotenv_settings,
            CaseInsensitivePyprojectTomlSource(
                settings_cls, pyproject_path if pyproject_path.exists() else None
            ),
            file_secret_settings,
        )

    add_security_headers: bool = True
    auth_providers: list[str] | None = []
    auth_required_roles: list[str] | None = []
    auth_entraid_client_id: str | None = None
    auth_entraid_tenant_id: str | None = None
    auth_entraid_multitenant: bool = False
    app_path: str = ""
    base_dir: Path = Path(__file__).resolve().parent
    object_cache_url: str | None = None
    object_cache_expire_at: str | None = "0 12 * * sat"
    object_cache_enable_compression: bool = True
    cors_allow_origins: list[str] = []
    custom_functions_max_retries: int = 3
    custom_functions_retry_codes: list[int] = [500, 502, 504]
    custom_headers: dict[str, str] = {}
    date_format: str | None = None
    taskpane_html: str = "taskpane.html"
    enable_alpinejs_csp: bool = True
    enable_bootstrap: bool = True
    enable_examples: bool = False
    enable_excel_online: bool = True
    enable_hotreload: bool = True
    enable_htmx: bool = True
    enable_socketio: bool = True
    enable_tests: bool = False
    enable_wasm: bool = False
    environment: Literal["dev", "qa", "uat", "staging", "prod"] = "prod"
    functions_namespace: str = "XLWINGS"
    hostname: str | None = None
    is_official_lite_addin: bool | None = False
    cdn_pyodide: bool = True
    cdn_officejs: bool = False
    log_level: str = "INFO"
    # Manifest UUIDs - loaded from pyproject.toml [tool.xlwings_server] or defaults
    # Run 'xlwings-server init' to generate unique UUIDs in pyproject.toml
    manifest_id_dev: UUID4 = "0a856eb1-91ab-4f38-b757-23fbe1f73130"
    manifest_id_qa: UUID4 = "9cda34b1-af68-4dc6-b97c-e63ef6284671"
    manifest_id_uat: UUID4 = "70428e53-8113-421c-8fe2-9b74fcb94ee5"
    manifest_id_staging: UUID4 = "34041f4f-9cb4-4830-afb5-db44b2a70e0e"
    manifest_id_prod: UUID4 = "4f342d85-3a49-41cb-90a5-37b1f2219040"
    project_name: str = "xlwings Server"
    public_addin_store: bool | None = None  # Deprecated. Use cdn_officejs instead.
    request_timeout: int | None = 300  # in seconds
    secret_key: str | None = None
    socketio_message_queue_url: str | None = None
    socketio_server_app: bool = False
    static_url_path: str = "/static"
    license_key: str | None = ""
    xlwings_version: str = get_package_version("xlwings")

    @computed_field
    @property
    def jsconfig(self) -> dict:
        return {
            "appPath": self.app_path,
            "authProviders": self.auth_providers,
            "customFunctionsMaxRetries": self.custom_functions_max_retries,
            "customFunctionsRetryCodes": self.custom_functions_retry_codes,
            "environment": self.environment,
            "isOfficialLiteAddin": self.is_official_lite_addin,
            "onWasm": self.enable_wasm,
            "requestTimeout": self.request_timeout,
            "xlwingsVersion": self.xlwings_version,
        }

    @computed_field
    @property
    def project_dir(self) -> Path:
        """Project directory - location of user's custom files"""
        return PROJECT_DIR

    @computed_field
    @property
    def package_dir(self) -> Path:
        """Package directory - location of xlwings-server package files"""
        return PACKAGE_DIR


# Try to import Settings from project directory (user override)
# Fall back to package Settings if not found
def _create_settings() -> Settings:
    """Create settings instance, checking for project-level config override"""
    try:
        config_file = PROJECT_DIR / "config.py"
        if config_file.exists():
            # Import Settings from project config
            import importlib

            config_module = importlib.import_module("config")
            if hasattr(config_module, "Settings"):
                ProjectSettings = config_module.Settings
                logger.info("Loaded settings from project config.py")
                return ProjectSettings()
    except Exception as e:
        logger.debug(f"No project config.py found or failed to load: {e}")

    # No project config or import failed, use package Settings
    return Settings()


settings = _create_settings()

# TODO: refactor once xlwings offers a runtime config
if settings.license_key and not os.getenv("XLWINGS_LICENSE_KEY"):
    os.environ["XLWINGS_LICENSE_KEY"] = settings.license_key

if settings.date_format:
    os.environ["XLWINGS_DATE_FORMAT"] = settings.date_format



from pathlib import Path

try:
    from ._version import __version__
except ImportError:
    __version__ = "0.0.0"

# Store package directory for use throughout the application
# This is evaluated when the package is first imported, before any sys.path manipulation
PACKAGE_DIR = Path(__file__).parent.resolve()

from .config import settings  # noqa: E402



import logging

from fastapi import APIRouter

from ..config import settings

router = APIRouter(prefix=settings.app_path)

logger = logging.getLogger(__name__)


# Endpoints
@router.get("/")
async def root():
    # This endpoint could be used for a health check
    return {"status": "ok"}



from .framework import Serializer, custom_decoder


class DefaultSerializer(Serializer):
    name = "default"

    @classmethod
    def serialize(cls, obj):
        return {
            "data": obj,
            "serializer": cls.name,
        }

    @classmethod
    def deserialize(cls, payload):
        return custom_decoder(payload["data"])


DefaultSerializer.register()



import argparse
import logging
import os
import shutil
import sys
import uuid
import zipfile
from datetime import datetime
from pathlib import Path
from textwrap import dedent
from urllib.parse import urljoin, urlparse

import uvicorn

from xlwings_server.build_utils import StaticFileHasher
from xlwings_server.config import PACKAGE_DIR, PROJECT_DIR


# Helper Classes and Functions
class FileTracker:
    """Track created and skipped files during CLI operations."""

    def __init__(self):
        self.created = []
        self.skipped = []

    def mark_created(self, path: str):
        """Mark a file as created."""
        self.created.append(path)

    def mark_skipped(self, path: str):
        """Mark a file as skipped (already exists)."""
        self.skipped.append(path)

    def print_summary(self, operation_name: str):
        """Print summary of created and skipped files."""
        print(f"\n{operation_name} complete!")


def validate_project_directory() -> Path:
    """Validate we're in an xlwings-server project directory."""
    project_path = Path.cwd()
    if not (project_path / "custom_functions").exists():
        print("Error: Not in an xlwings-server project directory.")
        print("Run this command from your project root.")
        print("Hint: Initialize a project first with 'xlwings-server init'")
        sys.exit(1)
    return project_path


def copy_file_if_not_exists(
    source: Path,
    dest: Path,
    tracker: FileTracker,
    display_name: str | None = None,
) -> bool:
    """Copy file if it doesn't exist. Returns True if copied."""
    display_name = display_name or str(dest)
    if dest.exists():
        tracker.mark_skipped(display_name)
        return False
    else:
        dest.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy(source, dest)
        tracker.mark_created(display_name)
        return True


def create_sample_file_if_not_exists(file_path: Path, content: str) -> bool:
    """Create a sample file if it doesn't exist. Returns True if created."""
    if file_path.exists():
        return False
    file_path.write_text(dedent(content))
    return True


def _ignore_dirs(directory: str, files: list[str]) -> set[str]:  # noqa: ARG001
    """Ignore function for shutil.copytree to skip .claude and __pycache__ dirs."""
    return {f for f in files if f in (".claude", "__pycache__")}


def _copy_folder(source_dir: Path, dest_dir: Path, folder_name: str) -> None:
    """Copy folder, replacing destination if it exists."""
    if source_dir.exists():
        if dest_dir.exists():
            shutil.rmtree(dest_dir)
        shutil.copytree(source_dir, dest_dir, ignore=_ignore_dirs)
        print(f"Copied {folder_name}.")
    else:
        print(f"No {folder_name} folder found.")


def _copy_folder_merge(source_dir: Path, dest_dir: Path, folder_name: str) -> None:
    """Copy folder contents, merging with existing files (overwrites on conflict)."""
    if source_dir.exists():
        shutil.copytree(source_dir, dest_dir, dirs_exist_ok=True, ignore=_ignore_dirs)
        print(f"Merged {folder_name}.")


# Requirements Parsing Helper Functions
def parse_requirements_in(file_path: Path) -> list[dict[str, str | None]]:
    """Parse requirements.in to extract package specifications.

    Returns list of dicts with keys:
    - name: package name (without extras or version)
    - extras: extras like 'linkify' or None
    - version_spec: version specifier from requirements.in or None
    """
    import re

    packages = []
    content = file_path.read_text()

    for line in content.splitlines():
        line = line.strip()
        # Skip empty lines, comments, and -r references
        if not line or line.startswith("#") or line.startswith("-r"):
            continue

        # Parse the package specification
        # Pattern: name[extras]version_spec ; markers
        # Extract name (before any of: [ < > = ! ~ ;)
        match = re.match(r"^([a-zA-Z0-9][-a-zA-Z0-9_.]*)", line)
        if not match:
            continue

        name = match.group(1)
        rest = line[len(name) :]

        # Extract extras if present
        extras = None
        extras_match = re.match(r"^\[([^\]]+)\]", rest)
        if extras_match:
            extras = extras_match.group(1)
            rest = rest[len(extras_match.group(0)) :]

        # Extract version specifier if present (before any marker)
        version_spec = None
        # Remove markers first (everything after ;)
        if ";" in rest:
            rest = rest.split(";")[0].strip()
        # Check for version specifier
        if rest and re.match(r"^[<>=!~]", rest):
            version_spec = rest.strip()

        packages.append(
            {
                "name": name,
                "extras": extras,
                "version_spec": version_spec,
            }
        )

    return packages


def parse_requirements_txt(file_path: Path) -> dict[str, str]:
    """Parse requirements.txt (lock file) to extract pinned versions.

    Returns dict mapping normalized_name -> version string.
    """
    import re

    from packaging.utils import canonicalize_name

    version_map = {}
    content = file_path.read_text()

    for line in content.splitlines():
        line = line.strip()
        # Skip empty lines, comments, and indented annotations
        if not line or line.startswith("#") or line.startswith(" "):
            continue

        # Look for pinned versions (package==version)
        if "==" not in line:
            continue

        # Remove markers (everything after ;)
        if ";" in line:
            line = line.split(";")[0].strip()

        # Split on == to get package and version
        parts = line.split("==", 1)
        if len(parts) != 2:
            continue

        package_part = parts[0].strip()
        version = parts[1].strip()

        # Extract package name (remove extras if present)
        match = re.match(r"^([a-zA-Z0-9][-a-zA-Z0-9_.]*)", package_part)
        if match:
            name = match.group(1)
            version_map[canonicalize_name(name)] = version

    return version_map


def extract_package_name_from_dependency(dep: str) -> str:
    """Extract package name from a dependency string.

    Examples:
        'foo>=1.0.0' -> 'foo'
        'foo[extra]>=1.0.0' -> 'foo'
        'foo>=1.0.0 ; sys_platform != "win32"' -> 'foo'
    """
    import re

    match = re.match(r"^([a-zA-Z0-9][-a-zA-Z0-9_.]*)", dep)
    return match.group(1) if match else dep


def relax_version_constraints(package_names: list[str]) -> None:
    """Replace == with >= for specified packages in pyproject.toml."""
    import tomlkit
    from packaging.utils import canonicalize_name

    pyproject_path = Path("pyproject.toml")
    if not pyproject_path.exists():
        return

    content = tomlkit.parse(pyproject_path.read_text())

    # Check for dependencies in [project.dependencies]
    if "project" not in content or "dependencies" not in content["project"]:
        return

    dependencies = content["project"]["dependencies"]

    # Create a set of normalized names for fast lookup
    names_to_relax = set(package_names)

    for i, dep in enumerate(dependencies):
        # Parse the dependency string to extract package name
        dep_name = extract_package_name_from_dependency(dep)
        if canonicalize_name(dep_name) in names_to_relax:
            # Replace == with >=
            if "==" in dep:
                new_dep = dep.replace("==", ">=", 1)
                dependencies[i] = new_dep
                print(f"  Relaxed version constraint: {dep} -> {new_dep}")

    pyproject_path.write_text(tomlkit.dumps(content))


# Migration Helper Functions
def validate_old_project_directory(old_path: Path) -> Path:
    """Validate that old_path contains a valid pre-1.0 project structure"""
    if not old_path.exists():
        print(f"Error: Path does not exist: {old_path}")
        sys.exit(1)

    if not old_path.is_dir():
        print(f"Error: Path is not a directory: {old_path}")
        sys.exit(1)

    # Check for app/ directory
    app_dir = old_path / "app"
    if not app_dir.exists():
        print("Error: Not an old xlwings-server project (no app/ directory found)")
        print(f"Expected to find: {app_dir}")
        sys.exit(1)

    # Check for app/config.py
    config_file = app_dir / "config.py"
    if not config_file.exists():
        print("Error: Not an old xlwings-server project (no app/config.py found)")
        print(f"Expected to find: {config_file}")
        sys.exit(1)

    # Check for app/custom_functions and app/custom_scripts
    if not (app_dir / "custom_functions").exists():
        print("Error: app/custom_functions/ directory not found")
        sys.exit(1)

    if not (app_dir / "custom_scripts").exists():
        print("Error: app/custom_scripts/ directory not found")
        sys.exit(1)

    return old_path


def extract_uuids_from_old_config(config_path: Path) -> dict[str, str]:
    """Extract manifest UUIDs from app/config.py using regex"""
    import re

    content = config_path.read_text()
    uuids = {}

    # Pattern matches both styles:
    # manifest_id_dev = "uuid-here"
    # manifest_id_dev: UUID4 = "uuid-here"
    pattern = r'manifest_id_(\w+)\s*[:=]\s*(?:UUID4\s*=\s*)?["\']([0-9a-f\-]{36})["\']'

    for match in re.finditer(pattern, content, re.MULTILINE):
        env_name, uuid_val = match.groups()
        uuids[f"manifest_id_{env_name}"] = uuid_val

    return uuids


def update_uuids_in_pyproject(uuids: dict[str, str]):
    """Update pyproject.toml with extracted UUIDs using tomlkit"""
    import tomlkit

    pyproject_path = Path("pyproject.toml")

    if not pyproject_path.exists():
        print("Error: pyproject.toml not found in current directory")
        sys.exit(1)

    content = tomlkit.parse(pyproject_path.read_text())

    # Ensure [tool.xlwings_server] section exists
    if "tool" not in content:
        content["tool"] = {}
    if "xlwings_server" not in content["tool"]:
        content["tool"]["xlwings_server"] = {}

    # Update UUIDs
    for key, value in uuids.items():
        content["tool"]["xlwings_server"][key] = value

    pyproject_path.write_text(tomlkit.dumps(content))


def is_file_customized(file_path: Path) -> bool:
    """Check if file exists and is not empty"""
    if not file_path.exists():
        return False

    # Check if file has content (not just whitespace)
    content = file_path.read_text().strip()
    return len(content) > 0


def copy_directory_recursive(
    source: Path,
    dest: Path,
    tracker: FileTracker,
    exclude_patterns: list[str] | None = None,
):
    """Recursively copy directory contents, tracking all files"""
    import shutil

    if not source.exists():
        return

    exclude_patterns = exclude_patterns or []

    for item in source.rglob("*"):
        if item.is_file():
            # Check if file matches any exclude pattern
            relative_path = item.relative_to(source)
            should_exclude = any(
                pattern in str(relative_path) for pattern in exclude_patterns
            )

            if should_exclude:
                continue

            dest_file = dest / relative_path

            # Create parent directories
            dest_file.parent.mkdir(parents=True, exist_ok=True)

            # Copy file
            shutil.copy2(item, dest_file)
            tracker.mark_created(str(dest / relative_path))


# Project Setup Functions
def create_project_structure(project_path: Path):
    """Create minimal project structure"""
    # Create project directory if it doesn't exist
    project_path.mkdir(parents=True, exist_ok=True)

    # Create required directories
    (project_path / "custom_functions").mkdir(exist_ok=True)
    (project_path / "custom_scripts").mkdir(exist_ok=True)
    (project_path / "templates").mkdir(exist_ok=True)
    (project_path / "certs").mkdir(exist_ok=True)

    # Create __init__.py files with imports
    functions_init = project_path / "custom_functions" / "__init__.py"
    if not functions_init.exists():
        functions_init.write_text(
            "# Import your custom functions here\nfrom .functions import *\n"
        )

    scripts_init = project_path / "custom_scripts" / "__init__.py"
    if not scripts_init.exists():
        scripts_init.write_text(
            "# Import your custom scripts here\nfrom .scripts import *\n"
        )

    # Create sample files
    create_sample_functions(project_path)
    create_sample_scripts(project_path)
    create_sample_taskpane(project_path)


def create_sample_functions(project_path: Path):
    """Create sample functions.py file with hello function"""
    functions_file = project_path / "custom_functions" / "functions.py"
    sample_code = """\
        from xlwings import func


        @func
        def hello(name):
            return f"Hello {name}!"
    """
    create_sample_file_if_not_exists(functions_file, sample_code)


def create_sample_scripts(project_path: Path):
    """Create sample scripts.py file with hello_world script"""
    scripts_file = project_path / "custom_scripts" / "scripts.py"
    sample_code = """\
        import xlwings as xw
        from xlwings import script


        @script
        def hello_world(book: xw.Book):
            sheet = book.sheets.active
            cell = sheet["A1"]
            if cell.value == "Hello xlwings!":
                cell.value = "Bye xlwings!"
            else:
                cell.value = "Hello xlwings!"
    """
    create_sample_file_if_not_exists(scripts_file, sample_code)


def create_sample_taskpane(project_path: Path):
    """Create sample taskpane.html template"""
    taskpane_file = project_path / "templates" / "taskpane.html"
    sample_html = """\
        {% extends "base.html" %}

        {% block content %}
        <div class="container-fluid pt-3 ps-3">
            <h1>{{ settings.project_name }}</h1>
            <button xw-click="hello_world" class="btn btn-primary btn-sm" type="button">
            Write 'Hello/Bye xlwings!' to A1
            </button>
        </div>
        {% endblock content %}
    """
    create_sample_file_if_not_exists(taskpane_file, sample_html)


def add_router_command():
    """Add router directory and sample router to project"""
    project_path = validate_project_directory()
    tracker = FileTracker()

    # Create routers directory
    routers_dir = project_path / "routers"
    routers_dir.mkdir(exist_ok=True)

    # Create __init__.py (empty)
    init_file = routers_dir / "__init__.py"
    if not init_file.exists():
        init_file.write_text("")

    # Create sample custom.py
    sample_file = routers_dir / "custom.py"
    sample_code = """\
        from fastapi import APIRouter, Request

        from xlwings_server import settings
        from xlwings_server.templates import TemplateResponse

        router = APIRouter(prefix=settings.app_path)


        @router.get("/hello-json")
        async def hello_json():
            return {"message": "Hello from custom router!"}


        @router.get("/hello-template")
        async def hello_template(request: Request):
            return TemplateResponse(
                request=request,
                name="examples/hello_world/taskpane_hello.html",
            )

    """

    if sample_file.exists():
        tracker.mark_skipped("custom.py")
    else:
        sample_file.write_text(dedent(sample_code))
        tracker.mark_created("custom.py")

    tracker.print_summary("Router setup")


def add_model_user_command():
    """Add user model to project for customization"""
    project_path = validate_project_directory()
    tracker = FileTracker()

    # Create models directory
    models_dir = project_path / "models"
    models_dir.mkdir(exist_ok=True)

    # Create __init__.py
    init_file = models_dir / "__init__.py"
    if not init_file.exists():
        init_file.write_text("from .user import User\n")

    # Copy user.py from package
    source_file = PACKAGE_DIR / "models" / "user.py"
    dest_file = models_dir / "user.py"
    copy_file_if_not_exists(source_file, dest_file, tracker, "user.py")

    tracker.print_summary("User model setup")


def add_auth_custom_command():
    """Add custom auth provider to project for customization"""
    import json

    from dotenv import dotenv_values, set_key

    project_path = validate_project_directory()
    tracker = FileTracker()

    # Create auth/custom directories
    auth_dir = project_path / "auth"
    custom_dir = auth_dir / "custom"
    custom_dir.mkdir(parents=True, exist_ok=True)

    # Create __init__.py files
    auth_init = auth_dir / "__init__.py"
    if not auth_init.exists():
        auth_init.write_text("")

    # Copy custom auth __init__.py from package
    source_file = PACKAGE_DIR / "auth" / "custom" / "__init__.py"
    dest_file = custom_dir / "__init__.py"
    copy_file_if_not_exists(source_file, dest_file, tracker, "auth/custom/__init__.py")

    # Create static/js/auth.js for custom auth
    static_js_dir = project_path / "static" / "js"
    static_js_dir.mkdir(parents=True, exist_ok=True)

    auth_js_dest = static_js_dir / "auth.js"
    auth_js_content = """\
        globalThis.getAuth = async function () {
          return {
            token: "test-token",  // TODO: implement
            provider: "custom",
          };
        };
    """

    if auth_js_dest.exists():
        tracker.mark_skipped("static/js/auth.js")
    else:
        auth_js_dest.write_text(dedent(auth_js_content))
        tracker.mark_created("static/js/auth.js")

    # Update .env to add "custom" to XLWINGS_AUTH_PROVIDERS
    env_file = project_path / ".env"
    if env_file.exists():
        env_content = env_file.read_text()

        # Check if XLWINGS_AUTH_PROVIDERS is commented
        if "# XLWINGS_AUTH_PROVIDERS=[]" in env_content:
            # Uncomment and set to ["custom"]
            env_content = env_content.replace(
                "# XLWINGS_AUTH_PROVIDERS=[]", 'XLWINGS_AUTH_PROVIDERS=["custom"]'
            )
            env_file.write_text(env_content)
            tracker.mark_created(".env (updated XLWINGS_AUTH_PROVIDERS)")
        elif "XLWINGS_AUTH_PROVIDERS=" in env_content:
            # Line is already uncommented, parse and update
            env_values = dotenv_values(env_file)
            current_providers = env_values.get("XLWINGS_AUTH_PROVIDERS", "")

            # Parse as JSON array (e.g., ["entraid"])
            if current_providers:
                try:
                    providers = json.loads(current_providers)
                except json.JSONDecodeError:
                    # Fallback to comma-separated if not JSON
                    providers = [p.strip() for p in current_providers.split(",")]
            else:
                providers = []

            # Add "custom" if not already present
            if "custom" not in providers:
                providers.append("custom")
                set_key(
                    env_file,
                    "XLWINGS_AUTH_PROVIDERS",
                    json.dumps(providers),
                    quote_mode="never",
                )
                tracker.mark_created(".env (updated XLWINGS_AUTH_PROVIDERS)")
            else:
                tracker.mark_skipped(".env (custom already in XLWINGS_AUTH_PROVIDERS)")
        else:
            tracker.mark_skipped(".env (XLWINGS_AUTH_PROVIDERS not found)")

    tracker.print_summary("Custom auth provider setup")


def add_config_command():
    """Add config.py to project for customizing settings"""
    project_path = validate_project_directory()
    tracker = FileTracker()

    # Check if config.py already exists
    config_file = project_path / "config.py"
    if config_file.exists():
        tracker.mark_skipped("config.py (already exists)")
    else:
        # Create config.py template
        template = dedent('''\
            """This uses pydantic-settings:
            https://docs.pydantic.dev/latest/concepts/pydantic_settings/
            """

            from xlwings_server.config import Settings as BaseSettings


            class Settings(BaseSettings):
                my_custom_setting: str = "default_value"

        ''')

        config_file.write_text(template)
        tracker.mark_created("config.py")

    tracker.print_summary("Config setup")


def add_auth_entraid_command():
    """Add Entra ID auth provider jwks.py to project for customization"""
    import json

    from dotenv import dotenv_values, set_key

    project_path = validate_project_directory()
    tracker = FileTracker()

    # Create auth/entraid directories
    auth_dir = project_path / "auth"
    entraid_dir = auth_dir / "entraid"
    entraid_dir.mkdir(parents=True, exist_ok=True)

    # Create __init__.py files
    auth_init = auth_dir / "__init__.py"
    if not auth_init.exists():
        auth_init.write_text("")

    # Copy jwks.py from package
    source_file = PACKAGE_DIR / "auth" / "entraid" / "jwks.py"
    dest_file = entraid_dir / "jwks.py"
    copy_file_if_not_exists(source_file, dest_file, tracker, "auth/entraid/jwks.py")

    # Update .env to add "entraid" to XLWINGS_AUTH_PROVIDERS
    env_file = project_path / ".env"
    if env_file.exists():
        env_content = env_file.read_text()

        # Check if XLWINGS_AUTH_PROVIDERS is commented
        if "# XLWINGS_AUTH_PROVIDERS=[]" in env_content:
            # Uncomment and set to ["entraid"]
            env_content = env_content.replace(
                "# XLWINGS_AUTH_PROVIDERS=[]", 'XLWINGS_AUTH_PROVIDERS=["entraid"]'
            )
            env_file.write_text(env_content)
            tracker.mark_created(".env (updated XLWINGS_AUTH_PROVIDERS)")
        elif "XLWINGS_AUTH_PROVIDERS=" in env_content:
            # Line is already uncommented, parse and update
            env_values = dotenv_values(env_file)
            current_providers = env_values.get("XLWINGS_AUTH_PROVIDERS", "")

            # Parse as JSON array (e.g., ["custom"])
            if current_providers:
                try:
                    providers = json.loads(current_providers)
                except json.JSONDecodeError:
                    # Fallback to comma-separated if not JSON
                    providers = [p.strip() for p in current_providers.split(",")]
            else:
                providers = []

            # Add "entraid" if not already present
            if "entraid" not in providers:
                providers.append("entraid")
                set_key(
                    env_file,
                    "XLWINGS_AUTH_PROVIDERS",
                    json.dumps(providers),
                    quote_mode="never",
                )
                tracker.mark_created(".env (updated XLWINGS_AUTH_PROVIDERS)")
            else:
                tracker.mark_skipped(".env (entraid already in XLWINGS_AUTH_PROVIDERS)")
        else:
            tracker.mark_skipped(".env (XLWINGS_AUTH_PROVIDERS not found)")

    tracker.print_summary("Entra ID auth provider setup")


def create_manifest_template(project_path: Path):
    """Copy manifest.xml template from package to project for customization"""
    source_file = PACKAGE_DIR / "templates" / "manifest.xml"
    dest_file = project_path / "templates" / "manifest.xml"

    if dest_file.exists():
        return

    # Create templates directory if it doesn't exist
    dest_file.parent.mkdir(parents=True, exist_ok=True)

    # Copy manifest template
    if source_file.exists():
        shutil.copy(source_file, dest_file)


def add_css_command(silent: bool = False):
    """Add style.css to project for customization"""
    project_path = validate_project_directory()
    tracker = FileTracker()

    source_file = PACKAGE_DIR / "static" / "css" / "style.css"
    dest_file = project_path / "static" / "css" / "style.css"

    copy_file_if_not_exists(source_file, dest_file, tracker, "static/css/style.css")
    if not silent:
        tracker.print_summary("CSS setup")


def add_js_command(silent: bool = False):
    """Add main.js to project for customization"""
    project_path = validate_project_directory()
    tracker = FileTracker()

    source_file = PACKAGE_DIR / "static" / "js" / "main.js"
    dest_file = project_path / "static" / "js" / "main.js"

    copy_file_if_not_exists(source_file, dest_file, tracker, "static/js/main.js")
    if not silent:
        tracker.print_summary("JS setup")


def create_dotenv(project_path: Path):
    """Copy .env.template from package to project as .env and set project name and secret key"""
    import secrets

    env_template_path = PACKAGE_DIR / ".env.template"
    env_path = project_path / ".env"

    if env_path.exists():
        return

    # Copy template
    shutil.copy(env_template_path, env_path)

    # Generate secret key
    secret_key = secrets.token_urlsafe(32)
    project_name = project_path.name

    # Read the .env file
    env_content = env_path.read_text()

    # Uncomment and set XLWINGS_PROJECT_NAME
    env_content = env_content.replace(
        '# XLWINGS_PROJECT_NAME=""', f'XLWINGS_PROJECT_NAME="{project_name}"'
    )

    # Uncomment and set XLWINGS_SECRET_KEY
    env_content = env_content.replace(
        'XLWINGS_SECRET_KEY=""', f'XLWINGS_SECRET_KEY="{secret_key}"'
    )

    # Write back
    env_path.write_text(env_content)


def create_gitignore(project_path: Path):
    """Create or update .gitignore with xlwings-server specific entries"""
    gitignore_path = project_path / ".gitignore"

    # xlwings-server specific entries
    xlwings_entries = [
        "",
        "# xlwings-server",
        ".env",
        "certs/",
        "*.pem",
    ]

    # Base Python entries (only used if .gitignore doesn't exist)
    base_entries = [
        "# Python-generated files",
        "__pycache__/",
        "*.py[oc]",
        "build/",
        "dist/",
        "wheels/",
        "*.egg-info",
        "",
        "# Virtual environments",
        ".venv",
    ]

    if gitignore_path.exists():
        # .gitignore exists - append only xlwings-server entries if not present
        content = gitignore_path.read_text()

        # Check if xlwings entries are already present
        has_xlwings_section = "# xlwings-server" in content

        if not has_xlwings_section:
            # Ensure file ends with newline before appending
            if content and not content.endswith("\n"):
                content += "\n"

            # Append xlwings-server specific entries
            content += "\n".join(xlwings_entries) + "\n"
            gitignore_path.write_text(content)
    else:
        # .gitignore doesn't exist - create with base + xlwings entries
        all_entries = base_entries + xlwings_entries
        gitignore_path.write_text("\n".join(all_entries) + "\n")


def create_uuids(project_path: Path | None = None):
    """Generate manifest UUIDs in pyproject.toml"""
    import tomlkit

    project_dir = project_path or Path.cwd()
    pyproject_path = project_dir / "pyproject.toml"

    if not pyproject_path.exists():
        # Create a minimal pyproject.toml
        data = tomlkit.document()
        data["tool"] = {}
        data["tool"]["xlwings_server"] = {}
    else:
        # Read existing pyproject.toml with tomlkit to preserve formatting
        content = pyproject_path.read_text()
        data = tomlkit.parse(content)

    # Check if UUIDs already exist
    if "tool" in data and "xlwings_server" in data["tool"]:
        existing_config = data["tool"]["xlwings_server"]
        if "manifest_id_dev" in existing_config:
            # Don't overwrite existing UUIDs
            return
    else:
        if "tool" not in data:
            data["tool"] = {}
        data["tool"]["xlwings_server"] = {}

    # Generate new UUIDs
    manifest_ids = {
        "manifest_id_dev": str(uuid.uuid4()),
        "manifest_id_qa": str(uuid.uuid4()),
        "manifest_id_uat": str(uuid.uuid4()),
        "manifest_id_staging": str(uuid.uuid4()),
        "manifest_id_prod": str(uuid.uuid4()),
    }

    # Add UUIDs to [tool.xlwings_server] section
    for key, value in manifest_ids.items():
        data["tool"]["xlwings_server"][key] = value

    # Write back (tomlkit preserves formatting)
    pyproject_path.write_text(tomlkit.dumps(data))


def init_command(path: str | None = None):
    """Initialize project"""
    # Determine project path
    if path is None:
        project_path = Path.cwd()

        # Ask for confirmation when initializing in current directory
        print(f"This will initialize an xlwings-server project in: {project_path}")
        response = input("Continue? [y/N]: ").strip().lower()
        if response not in ("y", "yes"):
            print("Initialization cancelled.")
            sys.exit(0)
    else:
        project_path = Path(path).resolve()

    create_project_structure(project_path)
    create_manifest_template(project_path)
    create_dotenv(project_path)
    create_gitignore(project_path)
    create_uuids(project_path)

    # Add CSS and JS files (reusing existing commands)
    original_cwd = Path.cwd()
    try:
        os.chdir(project_path)

        # Create empty images directory
        images_dir = project_path / "static" / "images"
        images_dir.mkdir(parents=True, exist_ok=True)

        # Add CSS and JS (silent mode to avoid cluttering init output)
        add_css_command(silent=True)
        add_js_command(silent=True)
    finally:
        os.chdir(original_cwd)

    print("Initialization complete!")


def add_azure_functions_command():
    """Add Azure Functions deployment files to project"""
    project_path = validate_project_directory()
    tracker = FileTracker()

    # Azure Functions template files are in xlwings_server/azure_functions_templates/
    source_dir = PACKAGE_DIR / "azure_functions_templates"

    azure_files = [
        "function_app.py",
        "host.json",
        ".funcignore",
        "local.settings.json",
    ]

    # Copy files with idempotency
    for filename in azure_files:
        source_file = source_dir / filename
        dest_file = project_path / filename

        if not source_file.exists():
            print(f"Warning: Source file not found: {source_file}")
            continue

        copy_file_if_not_exists(source_file, dest_file, tracker, filename)

    tracker.print_summary("Azure Functions setup")


def add_docker_command():
    """Add Docker deployment files to project"""
    project_path = validate_project_directory()
    tracker = FileTracker()

    source_dir = PACKAGE_DIR / "docker_templates"

    docker_files = [
        "Dockerfile",
        "docker-compose.yaml",
        ".dockerignore",
    ]

    for filename in docker_files:
        source_file = source_dir / filename
        dest_file = project_path / filename

        if not source_file.exists():
            print(f"Warning: Source file not found: {source_file}")
            continue

        copy_file_if_not_exists(source_file, dest_file, tracker, filename)

    tracker.print_summary("Docker setup")


def add_iis_command():
    """Add IIS deployment files to project"""
    import tomlkit

    project_path = validate_project_directory()
    tracker = FileTracker()

    web_config_content = dedent("""\
        <?xml version="1.0" encoding="UTF-8"?>
        <configuration>
          <system.webServer>
            <handlers>
              <add name="httpPlatformHandler" path="*" verb="*" modules="httpPlatformHandler" resourceType="Unspecified" />
            </handlers>
            <httpPlatform processPath="C:\\uv\\.venv\\Scripts\\python.exe"
                          arguments="-m uvicorn xlwings_server.main:main_app --host 127.0.0.1 --port %HTTP_PLATFORM_PORT%"
                          startupTimeLimit="60"
                          stdoutLogEnabled="true"
                          stdoutLogFile="C:\\inetpub\\xlwings-app\\logs\\stdout.log"
                          processesPerApplication="4">
              <environmentVariables>
                <environmentVariable name="PYTHONUNBUFFERED" value="1" />
                <environmentVariable name="XLWINGS_PROJECT_DIR" value="C:\\inetpub\\xlwings-app" />
              </environmentVariables>
            </httpPlatform>
          </system.webServer>
        </configuration>
        """)

    web_config_file = project_path / "web.config"
    if web_config_file.exists():
        tracker.mark_skipped("web.config")
    else:
        web_config_file.write_text(web_config_content)
        tracker.mark_created("web.config")

    # Update pyproject.toml with [tool.uv] exclude-dependencies
    pyproject_path = project_path / "pyproject.toml"
    if pyproject_path.exists():
        content = tomlkit.parse(pyproject_path.read_text())

        # Ensure [tool.uv] section exists
        if "tool" not in content:
            content["tool"] = {}
        if "uv" not in content["tool"]:
            content["tool"]["uv"] = {}

        # Check if exclude-dependencies already has pywin32
        existing_excludes = content["tool"]["uv"].get("exclude-dependencies", [])
        if "pywin32" not in existing_excludes:
            content["tool"]["uv"]["exclude-dependencies"] = existing_excludes + [
                "pywin32"
            ]
            pyproject_path.write_text(tomlkit.dumps(content))
            tracker.mark_created(
                "pyproject.toml (added [tool.uv] exclude-dependencies)"
            )
        else:
            tracker.mark_skipped("pyproject.toml ([tool.uv] already configured)")
    else:
        print("Warning: pyproject.toml not found, skipping uv configuration")

    tracker.print_summary("IIS setup")


def migrate_command(old_project_path: str):
    """Migrate from pre-1.0 project structure to new 1.0+ structure"""
    import shutil

    # Phase 1: Validation
    print("Starting migration...")

    # Validate current directory is a new project
    validate_project_directory()

    # Check for pyproject.toml
    if not Path("pyproject.toml").exists():
        print("Error: pyproject.toml not found in current directory")
        print("Run this command from an initialized xlwings-server 1.0+ project")
        sys.exit(1)

    # Validate old project structure
    old_path = validate_old_project_directory(Path(old_project_path))

    tracker = FileTracker()

    # Phase 2: Core File Migration
    # Remove default templates (from init) and old examples (from pre-1.0)
    files_to_remove = [
        Path("custom_functions/functions.py"),  # Default from xlwings-server init
        Path("custom_scripts/scripts.py"),  # Default from xlwings-server init
        Path("main.py"),  # Default from uv init
        Path("README.md"),  # Default from uv init
    ]
    for file_path in files_to_remove:
        if file_path.exists():
            file_path.unlink()

    # 1. Custom functions & scripts
    app_path = old_path / "app"
    copy_directory_recursive(
        app_path / "custom_functions",
        Path("custom_functions"),
        tracker,
        exclude_patterns=["__pycache__", "examples.py"],
    )
    copy_directory_recursive(
        app_path / "custom_scripts",
        Path("custom_scripts"),
        tracker,
        exclude_patterns=["__pycache__", "examples.py"],
    )

    # Clean up __init__.py files - only keep lines with "import *"
    for init_file in [
        Path("custom_functions/__init__.py"),
        Path("custom_scripts/__init__.py"),
    ]:
        if init_file.exists():
            content = init_file.read_text()
            lines = content.splitlines(keepends=True)

            # Only keep lines that contain "import *" (but not .examples)
            cleaned_lines = []
            for line in lines:
                # Keep only import * lines (excluding .examples imports)
                if "import *" in line and ".examples" not in line:
                    cleaned_lines.append(line)

            # Write cleaned content
            init_file.write_text("".join(cleaned_lines))

    # 2. Certificates
    old_certs = old_path / "certs"
    if old_certs.exists():
        copy_directory_recursive(old_certs, Path("certs"), tracker)

    # 3. .env file
    old_env = old_path / ".env"
    if old_env.exists():
        shutil.copy2(old_env, ".env")
        tracker.mark_created(".env")

    # 4. Templates - Manifest and Taskpane
    Path("templates").mkdir(exist_ok=True)

    # Copy manifest.xml as-is
    old_manifest = app_path / "templates" / "manifest.xml"
    if old_manifest.exists():
        shutil.copy2(old_manifest, "templates/manifest.xml")
        tracker.mark_created("templates/manifest.xml")

    # Determine which taskpane template was used in old project
    old_taskpane_router = app_path / "routers" / "taskpane.py"
    taskpane_template_name = None

    if old_taskpane_router.exists():
        # Parse the router to find the template name
        router_content = old_taskpane_router.read_text()

        # Look for template name when enable_examples is False
        import re

        # First check if there's conditional logic based on enable_examples
        if "enable_examples" in router_content:
            # Parse conditional: if not settings.enable_examples: name="..."
            match = re.search(
                r'if\s+not\s+settings\.enable_examples.*?name\s*=\s*["\']([^"\']+)["\']',
                router_content,
                re.DOTALL,
            )
            if match:
                taskpane_template_name = match.group(1)

        # If no conditional found, look for direct template name
        if not taskpane_template_name:
            # Look for: name="..." or name='...'
            match = re.search(r'name\s*=\s*["\']([^"\']+)["\']', router_content)
            if match:
                taskpane_template_name = match.group(1)
            else:
                # Look for: name=settings.something
                match = re.search(r"name\s*=\s*settings\.(\w+)", router_content)
                if match:
                    taskpane_template_name = "taskpane.html"  # Default fallback

    # Copy the identified taskpane template
    if taskpane_template_name:
        old_taskpane_template = app_path / "templates" / taskpane_template_name
        if old_taskpane_template.exists():
            shutil.copy2(old_taskpane_template, "templates/taskpane.html")
            tracker.mark_created(
                f"templates/taskpane.html (from {taskpane_template_name})"
            )
        else:
            # Template doesn't exist, try direct copy
            direct_taskpane = app_path / "templates" / "taskpane.html"
            if direct_taskpane.exists():
                shutil.copy2(direct_taskpane, "templates/taskpane.html")
                tracker.mark_created("templates/taskpane.html")
    else:
        # No router found, copy taskpane.html if it exists
        direct_taskpane = app_path / "templates" / "taskpane.html"
        if direct_taskpane.exists():
            shutil.copy2(direct_taskpane, "templates/taskpane.html")
            tracker.mark_created("templates/taskpane.html")

    # 5. Extract and update UUIDs
    uuids = extract_uuids_from_old_config(app_path / "config.py")
    if uuids:
        update_uuids_in_pyproject(uuids)
        tracker.mark_created("pyproject.toml (UUIDs updated)")

    # Phase 3: Optional Components Detection & Migration
    # Static files - CSS
    old_css = app_path / "static" / "css" / "style.css"
    if is_file_customized(old_css):
        Path("static/css").mkdir(parents=True, exist_ok=True)
        shutil.copy2(old_css, "static/css/style.css")
        tracker.mark_created("static/css/style.css")

    # Static files - JS
    old_js = app_path / "static" / "js" / "main.js"
    if is_file_customized(old_js):
        Path("static/js").mkdir(parents=True, exist_ok=True)
        shutil.copy2(old_js, "static/js/main.js")
        tracker.mark_created("static/js/main.js")

    # Static files - ribbon.js
    old_ribbon_js = app_path / "static" / "js" / "ribbon.js"
    if is_file_customized(old_ribbon_js):
        Path("static/js").mkdir(parents=True, exist_ok=True)
        shutil.copy2(old_ribbon_js, "static/js/ribbon.js")
        tracker.mark_created("static/js/ribbon.js")

    # Static files - Images
    old_images = app_path / "static" / "images"
    if old_images.exists() and any(old_images.iterdir()):
        copy_directory_recursive(old_images, Path("static/images"), tracker)

    # Static files - Other static content
    old_static = app_path / "static"
    if old_static.exists():
        for item in old_static.iterdir():
            # Skip already handled directories
            if item.name in ["css", "js", "images", "vendor"]:
                continue
            if item.is_dir():
                copy_directory_recursive(item, Path("static") / item.name, tracker)
            elif item.is_file():
                Path("static").mkdir(exist_ok=True)
                shutil.copy2(item, Path("static") / item.name)
                tracker.mark_created(f"static/{item.name}")

    # Custom templates (excluding package templates and already-handled files)
    old_templates = app_path / "templates"
    # These are package templates that should not be migrated
    excluded_templates = {
        "manifest.xml",  # Already handled specially
        "taskpane.html",  # Already handled specially
        "_book.html",  # Package template
        "alert_base.html",  # Package template
        "base.html",  # Package template
        "xlwings_alert.html",  # Package template
        "examples",  # Package examples directory
    }

    if old_templates.exists():
        for item in old_templates.iterdir():
            if item.name in excluded_templates:
                continue
            if item.is_file():
                dest_file = Path("templates") / item.name
                shutil.copy2(item, dest_file)
                tracker.mark_created(f"templates/{item.name}")
            elif item.is_dir():
                copy_directory_recursive(item, Path("templates") / item.name, tracker)

    # Parse and install dependencies from old requirements.in
    old_requirements_in = old_path / "requirements.in"
    old_requirements_txt = old_path / "requirements.txt"

    if old_requirements_in.exists():
        print("\nInstalling dependencies from requirements.in...")

        # Parse requirements.in for package names
        packages = parse_requirements_in(old_requirements_in)

        # Parse requirements.txt for pinned versions (optional)
        version_map: dict[str, str] = {}
        if old_requirements_txt.exists():
            version_map = parse_requirements_txt(old_requirements_txt)
        else:
            print("Note: requirements.txt not found, installing without version pins")

        # Build package specifications for installation
        dependencies_to_install = []
        packages_installed_with_pin = []

        from packaging.utils import canonicalize_name

        for pkg in packages:
            name = pkg["name"]
            extras = pkg["extras"]
            base_ref = f"{name}[{extras}]" if extras else name
            normalized = canonicalize_name(name)

            if pkg["version_spec"]:
                # Use version from requirements.in as-is
                dependencies_to_install.append(f"{base_ref}{pkg['version_spec']}")
            elif normalized in version_map:
                # Use pinned version from requirements.txt
                version = version_map[normalized]
                dependencies_to_install.append(f"{base_ref}=={version}")
                packages_installed_with_pin.append(normalized)
            else:
                # Package not in requirements.txt - install without version
                dependencies_to_install.append(base_ref)
                print(
                    f"  Note: {name} not found in requirements.txt, installing latest"
                )

        # Install dependencies using uv add
        if dependencies_to_install:
            import subprocess

            try:
                cmd = ["uv", "add"] + dependencies_to_install
                subprocess.run(cmd, check=True, capture_output=True, text=True)
                print(f"Successfully added: {', '.join(dependencies_to_install)}")
                tracker.mark_created(
                    f"Dependencies: {', '.join(d.split('==')[0] for d in dependencies_to_install)}"
                )

                # Relax version constraints after successful installation
                if packages_installed_with_pin:
                    relax_version_constraints(packages_installed_with_pin)
            except subprocess.CalledProcessError as e:
                print(f"Warning: Failed to install some dependencies: {e}")
                print(
                    f"You can manually install them with: uv add {' '.join(dependencies_to_install)}"
                )
            except FileNotFoundError:
                print("Warning: 'uv' command not found.")
                print(
                    f"Please manually install dependencies: uv add {' '.join(dependencies_to_install)}"
                )

    # Phase 4: Report
    tracker.print_summary("Migration")


def run_server():
    """Start the xlwings-server development server"""
    # Get project directory (where user runs the command)
    project_dir = Path.cwd()

    # Validate project structure
    custom_functions_dir = project_dir / "custom_functions"
    custom_scripts_dir = project_dir / "custom_scripts"

    if not custom_functions_dir.exists() or not custom_scripts_dir.exists():
        print(
            "Error: Project must have custom_functions/ and custom_scripts/ directories"
        )
        sys.exit(1)

    # Set environment variable so app can find project directory
    # The sys.path manipulation will happen in main.py before importing user modules
    os.environ["XLWINGS_PROJECT_DIR"] = str(project_dir)

    # Copy over required settings to wasm .env
    # This is done before starting the server to ensure wasm has up-to-date settings
    from xlwings_server.config import settings  # noqa: E402

    wasm_dir = project_dir / "wasm"
    if settings.enable_wasm and wasm_dir.exists():
        env_file = wasm_dir / ".env"
        create_wasm_settings(settings, env_file)

    # Determine SSL certificate paths
    is_cloud = os.getenv("CODESPACES")
    ssl_keyfile_path = project_dir / "certs" / "localhost+2-key.pem"
    ssl_certfile_path = project_dir / "certs" / "localhost+2.pem"

    ssl_keyfile = (
        str(ssl_keyfile_path) if ssl_keyfile_path.exists() and not is_cloud else None
    )
    ssl_certfile = (
        str(ssl_certfile_path) if ssl_certfile_path.exists() and not is_cloud else None
    )

    if (ssl_keyfile is None or ssl_certfile is None) and not is_cloud:
        print(
            "NO SSL KEYFILE OR CERTFILE FOUND. RUNNING ON HTTP, NOT HTTPS!.\n"
            "THIS WILL ONLY WORK WITH VBA AND OFFICE SCRIPTS, BUT NOT WITH "
            "OFFICE.JS ADD-INS!"
        )

    # Start uvicorn server
    # Note: We don't import settings here to avoid caching the xlwings_server module
    # before sys.path manipulation
    uvicorn.run(
        "xlwings_server.main:main_app",
        host="127.0.0.1",
        port=8000,
        reload=True,
        reload_dirs=[str(project_dir)],
        reload_includes=[".env"],
        ssl_keyfile=ssl_keyfile,
        ssl_certfile=ssl_certfile,
    )


def create_wasm_settings(settings, env_file):
    settings_map = {
        "XLWINGS_LICENSE_KEY": f'"{settings.license_key}"',
        "XLWINGS_ENABLE_EXAMPLES": str(settings.enable_examples).lower(),
        "XLWINGS_ENVIRONMENT": settings.environment,
        "XLWINGS_ENABLE_TESTS": str(settings.enable_tests).lower(),
        "XLWINGS_FUNCTIONS_NAMESPACE": settings.functions_namespace,
        "XLWINGS_IS_OFFICIAL_LITE_ADDIN": str(settings.is_official_lite_addin).lower(),
    }

    for key, value in settings_map.items():
        update_wasm_settings(key, value, env_file)


def update_wasm_settings(key: str, value: str, env_file: Path):
    if env_file.exists():
        content = env_file.read_text().splitlines()
    else:
        content = []

    key_found = False
    for i, line in enumerate(content):
        if line.startswith(f"{key}="):
            content[i] = f"{key}={value}"
            key_found = True
            break

    if not key_found:
        content.append(f"{key}={value}")

    env_file.parent.mkdir(parents=True, exist_ok=True)
    env_file.write_text("\n".join(content) + "\n")


def build_static_command(output_dir: str = "./dist", clean: bool = False):
    """Build static files and templates for production deployment with hashed filenames."""
    project_path = validate_project_directory()
    output_path = Path(output_dir)

    if clean and output_path.exists():
        shutil.rmtree(output_path)
        print("Output directory cleaned.")

    output_path.mkdir(exist_ok=True)

    # Copy static: package first, then project overlays
    _copy_folder(PACKAGE_DIR / "static", output_path / "static", "static (package)")
    _copy_folder_merge(
        project_path / "static", output_path / "static", "static (project)"
    )

    # Copy templates: package first, then project overlays
    _copy_folder(
        PACKAGE_DIR / "templates", output_path / "templates", "templates (package)"
    )
    _copy_folder_merge(
        project_path / "templates", output_path / "templates", "templates (project)"
    )

    # Hash static files and update references in templates
    hasher = StaticFileHasher(
        static_dir=output_path / "static", templates_dir=output_path / "templates"
    )
    hasher.process_files()

    print(f"\nBuild complete: {output_path.resolve()}")


def build_wasm_command(
    url, output_dir, create_zip=False, clean=False, environment=None
):
    import xlwings
    import xlwings as xw

    logging.getLogger("httpx").setLevel(logging.WARNING)
    build_timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")

    # Settings overrides
    parsed = urlparse(url)
    base_url = f"{parsed.scheme}://{parsed.netloc}".rstrip("/")
    app_path = parsed.path.rstrip("/")

    # TODO: these env vars aren't respected anymore, probably because of the config
    # import at the top. Need to be set directly in the environment where this is
    # running, e.g., CI
    os.environ["XLWINGS_ENABLE_WASM"] = "true"
    os.environ["XLWINGS_ENABLE_SOCKETIO"] = "false"
    os.environ["XLWINGS_APP_PATH"] = app_path
    os.environ["XLWINGS_STATIC_URL_PATH"] = f"{app_path}/static"

    if environment:
        os.environ["XLWINGS_ENVIRONMENT"] = environment

    from fastapi.testclient import TestClient  # noqa: E402

    from xlwings_server.config import settings  # noqa: E402
    from xlwings_server.main import main_app  # noqa: E402

    # Make sure settings is up-to-date
    create_wasm_settings(settings=settings, env_file=PROJECT_DIR / "wasm" / ".env")

    # Take the license key from .env
    os.environ["XLWINGS_LICENSE_KEY"] = settings.license_key
    import xlwings.pro

    output_dir = Path(output_dir)
    output_dir.mkdir(exist_ok=True)

    # Clean output directory
    if clean:
        if output_dir.exists():
            for filename in os.listdir(output_dir):
                file_path = output_dir / filename
                if file_path.is_file():
                    file_path.unlink()
                elif file_path.is_dir():
                    shutil.rmtree(file_path)
            print("Output directory cleaned.")

    # Endpoints
    client = TestClient(main_app)

    route_paths = [
        "manifest.xml",
        "taskpane.html",  # TODO: cover all routes from taskpane.py
        "xlwings/custom-functions-meta.json",
        "xlwings/custom-functions-code.js",
        "xlwings/pyodide.json",
    ]

    base_path = f"{app_path}/" if app_path else "/"
    routes = [urljoin(base_path, path) for path in route_paths]

    for ix, route in enumerate(routes):
        response = client.get(route)
        if response.status_code == 200:
            content = response.text
            filename = Path(route_paths[ix])
            if filename.name == "manifest.xml":
                content = content.replace("http://testserver", base_url)

            file_path = output_dir / filename
            file_path.parent.mkdir(parents=True, exist_ok=True)
            file_path.write_text(content)
        else:
            print(f"Failed to fetch {route} (status code: {response.status_code})")

    # Index.html
    index_path = output_dir / "index.html"
    index_path.write_text(f"This is an xlwings Wasm app! ({build_timestamp})")

    print("Static site generation complete.")

    # Copy from PACKAGE_DIR first (base files)
    _copy_folder(PACKAGE_DIR / "static", output_dir / "static", "static (package)")
    _copy_folder(PACKAGE_DIR / "wasm", output_dir / "wasm", "wasm (package)")
    _copy_folder(
        PACKAGE_DIR / "custom_functions",
        output_dir / "custom_functions",
        "custom_functions (package)",
    )
    _copy_folder(
        PACKAGE_DIR / "custom_scripts",
        output_dir / "custom_scripts",
        "custom_scripts (package)",
    )

    # Copy from PROJECT_DIR to overwrite with user customizations
    _copy_folder_merge(
        PROJECT_DIR / "static", output_dir / "static", "static (project)"
    )
    _copy_folder_merge(PROJECT_DIR / "wasm", output_dir / "wasm", "wasm (project)")
    _copy_folder_merge(
        PROJECT_DIR / "custom_functions",
        output_dir / "custom_functions",
        "custom_functions (project)",
    )
    _copy_folder_merge(
        PROJECT_DIR / "custom_scripts",
        output_dir / "custom_scripts",
        "custom_scripts (project)",
    )

    # .env
    try:
        deploy_key = xlwings.pro.LicenseHandler.create_deploy_key()
    except xw.LicenseError:
        deploy_key = settings.license_key
    update_wasm_settings(
        "XLWINGS_LICENSE_KEY", deploy_key, output_dir / "wasm" / ".env"
    )
    if environment:
        update_wasm_settings(
            "XLWINGS_ENVIRONMENT", environment, output_dir / "wasm" / ".env"
        )

    # Remove unused libraries
    def remove_dir_if_exists(path: Path) -> None:
        if path.exists():
            shutil.rmtree(path)

    remove_dir_if_exists(output_dir / "static" / "vendor" / "socket.io")
    if not settings.enable_alpinejs_csp:
        remove_dir_if_exists(output_dir / "static" / "vendor" / "@alpinejs")
    if settings.cdn_officejs:
        remove_dir_if_exists(output_dir / "static" / "vendor" / "@microsoft")
    if not settings.enable_bootstrap:
        remove_dir_if_exists(output_dir / "static" / "vendor" / "bootstrap")
        remove_dir_if_exists(output_dir / "static" / "vendor" / "bootstrap-xlwings")
    if not settings.enable_htmx:
        remove_dir_if_exists(output_dir / "static" / "vendor" / "htmx-ext-head-support")
        remove_dir_if_exists(
            output_dir / "static" / "vendor" / "htmx-ext-loading-states"
        )
        remove_dir_if_exists(output_dir / "static" / "vendor" / "htmx.org")

    def has_pyodide_requirement(requirements_file):
        if not requirements_file.exists():
            return False
        with open(requirements_file, "r") as f:
            return any("/static/vendor/pyodide/" in line for line in f)

    if settings.cdn_pyodide:
        requirements_path = output_dir / "wasm" / "requirements.txt"
        if not has_pyodide_requirement(requirements_path):
            remove_dir_if_exists(output_dir / "static" / "vendor" / "pyodide")

    # Remove unwanted files
    for cache_dir in (output_dir / "wasm").rglob("__pycache__"):
        remove_dir_if_exists(cache_dir)

    for ds_store in output_dir.rglob(".DS_Store"):
        ds_store.unlink(missing_ok=True)

    # Hash files
    hasher = StaticFileHasher(static_dir=output_dir, templates_dir=output_dir)
    hasher.process_files()

    # ZIP file
    if create_zip:
        zip_filename = output_dir / f"xlwings_wasm_{build_timestamp}.zip"

        try:
            with zipfile.ZipFile(zip_filename, "w", zipfile.ZIP_DEFLATED) as zipf:
                for file_path in output_dir.rglob("*"):
                    if file_path.is_file() and file_path != zip_filename:
                        arcname = file_path.relative_to(output_dir)
                        zipf.write(file_path, arcname)
            print(f"Created zip file: {zip_filename}")
        except Exception as e:
            print(f"Error creating zip file: {e}")


def main():
    """Entry point for xlwings-server CLI"""
    parser = argparse.ArgumentParser(
        description="xlwings Server - Modern Excel add-ins with Python"
    )
    subparsers = parser.add_subparsers(dest="command", help="Available commands")

    # Init command
    init_parser = subparsers.add_parser(
        "init",
        help="Initialize project",
    )
    init_parser.add_argument(
        "path",
        nargs="?",
        default=None,
        help="Project path (default: current directory). Use '.' for current directory or specify a path to create a new project.",
    )

    # Add command
    add_parser = subparsers.add_parser("add", help="Add optional components")
    add_subparsers = add_parser.add_subparsers(
        dest="add_category", help="Component categories"
    )

    # azure subcommand (with nested functions subcommand)
    azure_parser = add_subparsers.add_parser("azure", help="Azure integrations")
    azure_subparsers = azure_parser.add_subparsers(
        dest="azure_command", help="Azure services"
    )
    azure_subparsers.add_parser(
        "functions", help="Add Azure Functions deployment files"
    )

    # model subcommand (with nested user subcommand)
    model_parser = add_subparsers.add_parser("model", help="Data models")
    model_subparsers = model_parser.add_subparsers(
        dest="model_command", help="Model types"
    )
    model_subparsers.add_parser("user", help="Add user model for customization")

    # auth subcommand (with nested custom and entraid subcommands)
    auth_parser = add_subparsers.add_parser("auth", help="Authentication providers")
    auth_subparsers = auth_parser.add_subparsers(
        dest="auth_command", help="Auth provider types"
    )
    auth_subparsers.add_parser(
        "custom", help="Add custom auth provider for customization"
    )
    auth_subparsers.add_parser(
        "entraid", help="Add Entra ID auth provider jwks.py for customization"
    )

    # docker subcommand (standalone)
    add_subparsers.add_parser("docker", help="Add Docker deployment files")

    # router subcommand (standalone, no nesting needed)
    add_subparsers.add_parser("router", help="Add routers directory and sample router")

    # css subcommand (standalone)
    add_subparsers.add_parser("css", help="Add style.css for customization")

    # js subcommand (standalone)
    add_subparsers.add_parser("js", help="Add main.js for customization")

    # config subcommand (standalone)
    add_subparsers.add_parser("config", help="Add config.py for extending settings")

    # iis subcommand (standalone)
    add_subparsers.add_parser("iis", help="Add IIS deployment files (web.config)")

    # Build command with subcommands
    build_parser = subparsers.add_parser("build", help="Build commands for deployment")
    build_subparsers = build_parser.add_subparsers(
        dest="build_command", help="Build targets"
    )

    # build static subcommand
    build_static_parser = build_subparsers.add_parser(
        "static", help="Build static files and templates with hashed filenames"
    )
    build_static_parser.add_argument(
        "-o",
        "--output-dir",
        help="Output directory path (default: ./dist)",
        type=str,
        default="./dist",
    )
    build_static_parser.add_argument(
        "--no-clean",
        help="Don't clean the output directory before building",
        action="store_true",
    )

    # build wasm subcommand
    build_wasm_parser = build_subparsers.add_parser(
        "wasm", help="Build xlwings Wasm distribution"
    )
    build_wasm_parser.add_argument(
        "url", help="URL of where the xlwings Wasm app will be hosted"
    )
    build_wasm_parser.add_argument(
        "-o",
        "--output-dir",
        help="Output directory path (default: ./dist)",
        type=str,
        default="./dist",
    )
    build_wasm_parser.add_argument(
        "-z",
        "--create-zip",
        help="Create zip archive in addition to the static files",
        action="store_true",
    )
    build_wasm_parser.add_argument(
        "--no-clean",
        help="Don't clean the output directory before building",
        action="store_true",
    )
    build_wasm_parser.add_argument(
        "-e",
        "--environment",
        help="Sets XLWINGS_ENVIRONMENT (default: value from .env)",
        type=str,
    )

    # Migrate command
    migrate_parser = subparsers.add_parser(
        "migrate", help="Migrate from pre-1.0 project structure"
    )
    migrate_parser.add_argument(
        "old_project_path",
        help="Path to old xlwings-server project (directory containing app/)",
    )

    args = parser.parse_args()

    if args.command == "init":
        init_command(args.path)
    elif args.command == "add":
        if args.add_category == "azure":
            if args.azure_command == "functions":
                add_azure_functions_command()
            else:
                print("Error: Please specify Azure service (e.g., functions)")
                sys.exit(1)
        elif args.add_category == "model":
            if args.model_command == "user":
                add_model_user_command()
            else:
                print("Error: Please specify model type (e.g., user)")
                sys.exit(1)
        elif args.add_category == "auth":
            if args.auth_command == "custom":
                add_auth_custom_command()
            elif args.auth_command == "entraid":
                add_auth_entraid_command()
            else:
                print("Error: Please specify auth provider (e.g., custom, entraid)")
                sys.exit(1)
        elif args.add_category == "docker":
            add_docker_command()
        elif args.add_category == "router":
            add_router_command()
        elif args.add_category == "css":
            add_css_command()
        elif args.add_category == "js":
            add_js_command()
        elif args.add_category == "config":
            add_config_command()
        elif args.add_category == "iis":
            add_iis_command()
        else:
            print("Error: Please specify what to add")
            print("Available: azure, docker, model, auth, router, css, js, config, iis")
            sys.exit(1)
    elif args.command == "build":
        if args.build_command == "static":
            build_static_command(output_dir=args.output_dir, clean=not args.no_clean)
        elif args.build_command == "wasm":
            build_wasm_command(
                url=args.url,
                output_dir=args.output_dir,
                create_zip=args.create_zip,
                clean=not args.no_clean,
                environment=args.environment,
            )
        else:
            print("Error: Please specify build target (e.g., static, wasm)")
            sys.exit(1)
    elif args.command == "migrate":
        migrate_command(args.old_project_path)
    else:
        # Default: run server
        run_server()


if __name__ == "__main__":
    main()



"""
Make sure to import the desired functions under __init__.py, e.g.:
from .example import *
"""

import asyncio
import datetime as dt
import sys
from pathlib import Path
from typing import Annotated

import numpy as np
import pandas as pd
from xlwings.server import arg, func, ret

from . import settings

if not settings.enable_wasm:
    # Try to import custom_scripts from project directory first (CLI/Azure mode)
    # Fall back to package location (tests/package mode)
    try:
        import custom_scripts
    except ModuleNotFoundError:
        import xlwings_server.custom_scripts as custom_scripts

    from xlwings.constants import ObjectHandleIcons
    from xlwings.ext.sql import _sql

    from xlwings_server import utils
    from xlwings_server.models import CurrentUser


# 1) This is the most basic custom function -- it only requires the @func decorator.
@func
def hello(name):
    return f"Hello {name}!"


# 2) Returning a pandas DataFrame and function documentation
# The function's doc string will turn up in the Function Wizard together with the
# doc strings from the arg decorators. The sample also shows how to suppress the index
# of a DataFrame via the ret decorator.
@func
@arg("rows", doc="The number of rows in the returned array.")
@arg("cols", doc="The number of columns in the returned array.")
def standard_normal(rows, cols):
    """Returns an array of standard normally distributed pseudo random numbers"""
    rng = np.random.default_rng()
    matrix = rng.standard_normal(size=(rows, cols))
    date_rng = pd.date_range(start=dt.datetime(2025, 6, 15), periods=rows, freq="D")
    df = pd.DataFrame(
        matrix, columns=[f"col{i+1}" for i in range(matrix.shape[1])], index=date_rng
    )
    return df


# 3) Reading a pandas DataFrames
@func
@arg("df", pd.DataFrame)
def correl(df):
    """Like CORREL, but it works on whole matrices instead of just 2 arrays."""
    return df.corr()


# 4) Type hints: this is the same example as 3), but using type hints instead of
# decorators. You could also use type hints and decorators together. In this sample, we
# are storing the Annotated type hint outside of the function, so it is easy to reuse.
Df = Annotated[pd.DataFrame, {"index": True}]


@func
def correl2(df: Df):
    """Like CORREL, but it works on whole matrices instead of just 2 arrays."""
    return df.corr()


if not settings.enable_wasm:
    # 5) Object handles: This returns an object handle to a DataFrame that is generated on
    # the backend. You can change the `text` and `icon` via annotated type hint or via ret
    # decorator.
    @func
    async def get_df() -> object:
        """Returns an object handle to the Excel cell (for production, this requires
        XLWINGS_OBJECT_CACHE_URL)."""
        return pd.DataFrame(
            {"A": [1, 2, 3, 4, 5], "B": [10, 8, 6, 4, 2], "C": [10, 9, 8, 7, 6]}
        )

    # 6) Object handles: demonstrating the use of the "icon" and "text" options. Instead
    # of using the ret decorator, you could also use an annotated type hint like so:
    # -> Annotated[object, {"icon": ObjectHandleIcons.table, "text": "healthexp"}]
    @func
    @ret(icon=ObjectHandleIcons.table, text="healthexp")
    async def get_healthexp(
        csv_url="https://raw.githubusercontent.com/mwaskom/seaborn-data/master/healthexp.csv",
    ) -> object:
        """Returns an object handle to the Excel cell (for production, this requires
        XLWINGS_OBJECT_CACHE_URL)."""
        return pd.read_csv(csv_url)

    # 7) Object handles: turn an Excel range into a DataFrame object handle. Using an Excel
    # table as source makes it easy to work with dynamic source ranges. This sample reuses
    # the Df type hint from above to set index=False.
    @func
    async def to_df(df: Df) -> object:
        return df

    # 8) Object handles: use a pandas DataFrame query by providing a DataFrame via object
    # handle and the query as string: [NAMESPACE].DF_QUERY(A1, "A > B")
    @func
    async def df_query(df: object, query: str) -> Df:
        return df.query(query)

    # 9) Object handles: Generic function that turns an object handle into Excel values
    @func
    async def view(obj: object, head=None):
        """Converts an object handle to cell values. `head` can be TRUE or an integer, which
        represents the number of rows from the top that you want to see. TRUE returns the
        first 5 rows.
        """
        if head and isinstance(head, bool):
            head = 5
        if isinstance(obj, pd.DataFrame) and head:
            return obj.iloc[:head, :]
        elif isinstance(obj, (list, tuple, np.ndarray)) and head:
            return obj[:head]
        else:
            return obj

    # 10) Object handles: Clear the object cache manually
    @func
    async def clear_object_cache():
        """Object handle: Clear the object cache manually"""
        await utils.clear_object_cache()
        return "Object cache cleared"

    # 11) Streaming functions (the modern version of RTD functions)
    @func
    async def streaming_random(rows, cols):
        """Streaming function: must be provided as async generator,
        requires XLWINGS_ENABLE_SOCKETIO=true
        """
        rng = np.random.default_rng()
        while True:
            matrix = rng.standard_normal(size=(rows, cols))
            df = pd.DataFrame(
                matrix, columns=[f"col{i+1}" for i in range(matrix.shape[1])]
            )
            yield df
            await asyncio.sleep(1)

    # 12) To access the current user object, simply add an argument with the CurrentUser
    # type hint
    @func
    def get_current_user(current_user: CurrentUser):
        return f"{current_user}"

    # 13) In-Excel SQL: query Excel ranges/tables via SQL (SQLite dialect)
    @func
    @arg("tables", expand="table", ndim=2)
    def sql(query, *tables):
        """In-Excel SQL
        see: https://docs.xlwings.org/en/latest/extensions.html#in-excel-sql"""
        return _sql(query, *tables)

    # 14) Custom functions with side effects (experimental)
    @func
    async def hello_with_script(name):
        """This function triggers a custom script, requires XLWINGS_ENABLE_SOCKETIO=true"""
        await utils.trigger_script(custom_scripts.hello_world, exclude="MySheet")
        return f"Hello {name}!"


# Unit tests
if settings.enable_tests:
    sys.path.append(str(Path(__file__).parent.parent.resolve()))
    from tests.e2e_custom_functions import *



import os

from dotenv import load_dotenv

load_dotenv(".env")

# xlwings depends on this, so set before importing xlwings
os.environ["XLWINGS_ON_SERVER"] = "true"


def read_bool(env_name: str, default: str = "false") -> bool:
    return os.getenv(env_name, default).lower() == "true"


class Settings:
    enable_examples: bool = read_bool("XLWINGS_ENABLE_EXAMPLES")
    enable_wasm: bool = True
    enable_tests: bool = read_bool("XLWINGS_ENABLE_TESTS")
    environment: str = os.getenv("XLWINGS_ENVIRONMENT", "prod")
    functions_namespace: str = os.getenv("XLWINGS_FUNCTIONS_NAMESPACE", "XLWINGS")
    is_official_lite_addin: bool = read_bool("XLWINGS_IS_OFFICIAL_LITE_ADDIN")


settings = Settings()



import logging

import redis

from . import object_handles
from .config import settings
from .routers import socketio as socketio_router, xlwings as xlwings_router

logger = logging.getLogger(__name__)


async def trigger_script(script, **options):
    sid = xlwings_router.socketio_id_context.get()
    if not isinstance(script, str):
        script = script.__name__
    await socketio_router.sio.emit(
        "xlwings:trigger-script",
        {"script_name": script, **options},
        to=sid,
    )
    logger.info(f"Script '{script}' triggered for sid '{sid}' with config: {options}")


async def clear_object_cache():
    if settings.object_cache_url:
        redis_client: redis.Redis = xlwings_router.redis_client_context.get()
        keys = redis_client.scan_iter(match="object:*")
        for key in keys:
            redis_client.delete(key)
        logger.info("Cleared all keys starting with 'object:' from the Redis cache")
    else:
        object_handles.cache = {}



import json

from . import (
    default_serializer,
    dictionary_serializer,
    numpy_serializer,
    pandas_serializer,
)
from .framework import Serializer, custom_encoder, serializers


def serialize(obj, serializer_name="default"):
    serializer = serializers.get(type(obj), serializers.get(serializer_name))
    if serializer is None:
        raise ValueError(f"No serializer registered for object of type {type(obj)}")
    return json.dumps(serializer.serialize(obj), default=custom_encoder)


def deserialize(payload, serializer_name="default"):
    payload = json.loads(payload)
    serializer_name = payload.get("serializer", serializer_name)
    serializer = serializers.get(serializer_name)
    if serializer is None:
        raise ValueError(f"No serializer registered with name '{serializer_name}'")
    return serializer.deserialize(payload)



"""
Azure Functions

Note: Azure Functions don't support streaming functions/socket.io.

The other files required for Azure Functions are:
- host.json
- local.settings.json
- .funcignore

The function is always called http_app_func, see:
https://github.com/Azure-Samples/fastapi-on-azure-functions/issues/31

For app logs, in Azure portal go to:
Function App > My Function App. Than, under `http_app_func`, click on `Invocations and more`.
"""

import os
from pathlib import Path

import azure.functions as func

# Must come before importing xlwings_server
os.environ["XLWINGS_PROJECT_DIR"] = str(Path(__file__).parent)

from xlwings_server.main import main_app  # noqa: E402

app = func.AsgiFunctionApp(app=main_app, http_auth_level=func.AuthLevel.ANONYMOUS)



import logging
import secrets

from aiocache import cached
from fastapi import status
from fastapi.exceptions import HTTPException

from xlwings_server.models import User

logger = logging.getLogger(__name__)


@cached(ttl=60 * 60)
async def validate_token(token_string: str):
    """token_string has to be delivered via the Authorization header from Excel. The
    Authorization header is set by implementing the globalThis.getAuth() function
    under static/js/auth.js.
    See https://server.xlwings.org/en/latest/auth_providers/#custom-authentication
    """
    if secrets.compare_digest(token_string, "test-token"):  # TODO: implement
        return User(id="customid", name="custom user")
    else:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Custom Auth Error: Couldn't validate token",
        )



from aiocache import cached


@cached(ttl=60 * 60 * 24)
async def get_jwks_json():
    """If your application runs on an air-gapped server and can't download the
    Azure JSON Web Key Set (JWKS), you can provide your own function here to access
    the JSON file.
    """
    pass



"""Static file hasher for cache-busting in production builds.

This module provides the StaticFileHasher class which adds content-based hashes
to static file names (e.g., main.js -> main.a1b2c3d4.js) and updates all references
in HTML, JS, and CSS files accordingly. This is primarily used during the Wasm
build process to enable effective cache-busting in production deployments.
"""

import hashlib
import os
import re
from pathlib import Path


class StaticFileHasher:
    """Hashes static files and updates references for cache-busting.

    This utility is used during the Wasm build process to add content-based
    hashes to static file names (e.g., main.js -> main.a1b2c3d4.js) and update
    all references in HTML, JS, and CSS files accordingly.

    Args:
        static_dir: Directory containing static files to hash
        templates_dir: Directory containing templates with references to update
    """

    def __init__(self, static_dir: Path, templates_dir: Path):
        self.static_dir = static_dir
        self.templates_dir = templates_dir
        self.file_mapping: dict[
            str, dict[str, str]
        ] = {}  # rel_path -> {old_name: new_name, path: Path}
        self.processed_files: set[Path] = set()

    def get_relative_path(self, path: Path, base_dir: Path) -> Path:
        """Get the relative path from base_dir to the given path"""
        try:
            return path.relative_to(base_dir)
        except ValueError:
            return path

    def should_process_file(self, path: Path) -> bool:
        """Determine if a file should be processed based on various criteria"""
        if not path.is_file() or path.name.startswith("."):
            return False

        excluded_patterns = {
            ".map",
            ".md",
            ".py",
            ".scss",
            ".txt",
            ".whl",
            ".xml",
            "custom-functions-code.js",
            "custom-functions-meta.json",
            "eula.html",
            "fonts/",
            "images/ribbon/",
            "index.html",
            "manifest.xml",
            "privacy.html",
            "support.html",
            "taskpane.html",
            "vendor/",
        }

        path_str = str(path)
        return (
            "." in path.name  # requires extension
            and not any(pattern in path_str for pattern in excluded_patterns)
        )

    def hash_file(self, path: Path) -> str:
        """Generate a hash for the given file"""
        contents = path.read_bytes()
        return hashlib.sha256(contents).hexdigest()[:8]

    def generate_new_name(self, filename: str, digest: str) -> str:
        """Generate the new filename with hash included"""
        name, ext = os.path.splitext(filename)
        return f"{name}.{digest}{ext}"

    def get_replacement_patterns(
        self, file_path: Path, rel_path: str, old_name: str, new_name: str
    ) -> list[tuple[str, str]]:
        """Generate all possible path patterns for replacement"""
        patterns = []
        dir_path = os.path.dirname(rel_path)

        # For absolute paths (mainly HTML references)
        abs_patterns = [
            (f"{rel_path}", f"{dir_path}/{new_name}"),
            (f"/{rel_path}", f"/{dir_path}/{new_name}"),
        ]

        # For relative paths (mainly JS imports)
        try:
            current_dir = file_path.parent.relative_to(self.static_dir)
            target_dir = Path(dir_path)
            rel_import_path = os.path.relpath(target_dir, current_dir)

            # Handle the case where files are in the same directory
            if rel_import_path == ".":
                rel_patterns = [
                    (f"./{old_name}", f"./{new_name}"),
                    (old_name, new_name),
                ]
            else:
                rel_import_path = rel_import_path.replace("\\", "/")
                if not rel_import_path.startswith("."):
                    rel_import_path = f"./{rel_import_path}"
                rel_patterns = [
                    (f"{rel_import_path}/{old_name}", f"{rel_import_path}/{new_name}"),
                ]
        except ValueError:
            rel_patterns = []

        # Combine all patterns
        all_paths = abs_patterns + rel_patterns

        # Generate variations for each pattern
        for old_path, new_path in all_paths:
            # Normalize slashes
            old_path = old_path.replace("\\", "/")
            new_path = new_path.replace("\\", "/")

            # Remove any double slashes
            old_path = re.sub(r"/{2,}", "/", old_path)
            new_path = re.sub(r"/{2,}", "/", new_path)

            variations = [
                (old_path, new_path),
                (f"'{old_path}'", f"'{new_path}'"),
                (f'"{old_path}"', f'"{new_path}"'),
                (f'from "{old_path}"', f'from "{new_path}"'),
                (f"from '{old_path}'", f"from '{new_path}'"),
                (f'import "{old_path}"', f'import "{new_path}"'),
                (f"import '{old_path}'", f"import '{new_path}'"),
                (f"url({old_path})", f"url({new_path})"),
                (f'url("{old_path}")', f'url("{new_path}")'),
                (f"url('{old_path}')", f"url('{new_path}')"),
            ]
            patterns.extend(variations)

        return patterns

    def replace_in_file(
        self, file_path: Path, excluded_dirs: list[str] | None = None
    ) -> None:
        """Replace all occurrences of original filenames with hashed versions"""
        if excluded_dirs is None:
            excluded_dirs = []

        if any(excluded_dir in file_path.parts for excluded_dir in excluded_dirs):
            return

        content = file_path.read_text()
        new_content = content

        # Sort paths by length (longest first) to avoid partial replacements
        sorted_paths = sorted(self.file_mapping.keys(), key=len, reverse=True)

        for rel_path in sorted_paths:
            file_info = self.file_mapping[rel_path]
            old_name = file_info["old_name"]
            new_name = file_info["new_name"]

            patterns = self.get_replacement_patterns(
                file_path, rel_path, old_name, new_name
            )

            for old_pattern, new_pattern in patterns:
                new_content = new_content.replace(old_pattern, new_pattern)

        if new_content != content:
            file_path.write_text(new_content)

    def process_files(self):
        """Process all static files and update references"""
        # First pass: hash all files and create mapping
        for source_path in self.static_dir.rglob("*"):
            if self.should_process_file(source_path):
                rel_path = self.get_relative_path(source_path, self.static_dir)
                rel_path_str = str(rel_path).replace("\\", "/")

                digest = self.hash_file(source_path)
                old_name = source_path.name
                new_name = self.generate_new_name(old_name, digest)

                self.file_mapping[rel_path_str] = {
                    "old_name": old_name,
                    "new_name": new_name,
                    "path": source_path,
                }

        # Second pass: rename files
        for file_info in self.file_mapping.values():
            old_path = file_info["path"]
            if old_path.exists():  # Check if not already renamed
                new_path = old_path.with_name(file_info["new_name"])
                old_path.rename(new_path)

        # Third pass: update references in files
        for template_file in self.templates_dir.rglob("*.html"):
            self.replace_in_file(template_file)

        for js_file in self.static_dir.rglob("*.js"):
            self.replace_in_file(js_file, excluded_dirs=["vendor"])

        for css_file in self.static_dir.rglob("*.css"):
            self.replace_in_file(css_file, excluded_dirs=["vendor"])



import logging

import socketio
import xlwings as xw

# Try to import custom modules from project directory first (CLI/Azure mode)
# Fall back to package location (tests/package mode)
try:
    import custom_functions
except ModuleNotFoundError:
    import xlwings_server.custom_functions as custom_functions

from xlwings_server.config import PROJECT_DIR, settings
from xlwings_server.dependencies import authenticate
from xlwings_server.models import CurrentUser

logger = logging.getLogger(__name__)

if settings.socketio_message_queue_url:
    client_manager = socketio.AsyncRedisManager(
        settings.socketio_message_queue_url, write_only=not settings.socketio_server_app
    )
else:
    client_manager = None


sio = socketio.AsyncServer(
    async_mode="asgi",
    client_manager=client_manager,
    cors_allowed_origins=(
        settings.cors_allow_origins[0]
        if len(settings.cors_allow_origins) == 1
        else settings.cors_allow_origins
    ),
)


@sio.on("connect")
async def connect(sid, environ, auth):
    if settings.environment == "dev" and settings.enable_hotreload:
        from .. import hotreload

        logging.getLogger("watchfiles").setLevel(logging.ERROR)
        await hotreload.start_browser_reload_watcher(sio=sio, directory=PROJECT_DIR)
    token_string = auth.get("token")
    provider = auth.get("provider")
    try:
        current_user = await authenticate(token_string, auth_provider=provider)
        await sio.save_session(sid, {"current_user": current_user})
        logger.info(f"Socket.io: connect {sid}")
        logger.info(f"Socket.io: User authenticated {current_user.name}")
    except Exception as e:
        logger.info(f"Socket.io: authentication failed for sid {sid}: {repr(e)}")
        await sio.disconnect(sid)


@sio.on("disconnect")
async def disconnect(sid):
    await xw.server.sio_disconnect(sid)


@sio.on("xlwings:function-call")
async def sio_function_call(sid, data):
    session = await sio.get_session(sid)
    current_user = session["current_user"]
    logger.info(f"""Function "{data['func_name']}" called by {current_user.name}""")
    await xw.server.sio_custom_function_call(
        sid, data, custom_functions, current_user, sio, {CurrentUser: current_user}
    )



import uuid
from io import StringIO

try:
    import pandas as pd
except ImportError:
    pd = None

from .framework import Serializer

if pd:

    class PandasDataFrameSerializer(Serializer):
        name = "pd.DataFrame"

        @classmethod
        def serialize(cls, df):
            serialized = {"serializer": cls.name}

            if isinstance(df.index, pd.MultiIndex):
                df = df.copy()
                index_names = df.index.names
                # Index names are often None
                temp_names = [
                    f"_idx_{uuid.uuid4().hex[:8]}" for _ in range(len(df.index.names))
                ]
                index_mapping = {
                    temp: orig for temp, orig in zip(temp_names, index_names)
                }
                df.index.names = temp_names
                df = df.reset_index()
                serialized.update(
                    {
                        "is_multi_index": True,
                        "index_mapping": index_mapping,
                    }
                )

            serialized.update(
                {
                    "serializer": cls.name,
                    "data": df.to_json(date_format="iso"),
                    "dtypes": {col: str(dtype) for col, dtype in df.dtypes.items()},
                }
            )

            # Preserve DatetimeIndex frequency if it exists
            if isinstance(df.index, pd.DatetimeIndex) and df.index.freq is not None:
                serialized["index_freq"] = df.index.freq.freqstr

            return serialized

        @classmethod
        def deserialize(cls, payload):
            df = pd.read_json(StringIO(payload["data"]))

            # Standard column type conversion
            for col, dtype in payload["dtypes"].items():
                df[col] = df[col].astype(dtype)

            # Handle MultiIndex reconstruction
            if payload.get("is_multi_index"):
                index_mapping = payload["index_mapping"]
                # Convert temporary columns back to index with original names
                index_cols = list(index_mapping.keys())
                df = df.set_index(index_cols)
                # Restore original index names
                df.index.names = list(index_mapping.values())

            # Restore DatetimeIndex frequency if it was saved
            if isinstance(df.index, pd.DatetimeIndex) and "index_freq" in payload:
                df.index.freq = pd.tseries.frequencies.to_offset(payload["index_freq"])

            return df

    PandasDataFrameSerializer.register(pd.DataFrame)

    class PandasSeriesSerializer(Serializer):
        name = "pd.Series"

        @classmethod
        def serialize(cls, series):
            return {
                "serializer": cls.name,
                "data": series.to_json(date_format="iso"),
                "dtype": str(series.dtype),
            }

        @classmethod
        def deserialize(cls, payload):
            series = pd.read_json(StringIO(payload["data"]), typ="series")
            series = series.astype(payload["dtype"])
            return series

    PandasSeriesSerializer.register(pd.Series)



try:
    import numpy as np
except ImportError:
    np = None

from .framework import Serializer

if np:

    class NumpyArraySerializer(Serializer):
        name = "np.array"

        @classmethod
        def serialize(cls, arr: np.ndarray):
            return {
                "serializer": cls.name,
                "data": arr.tolist(),
                "dtype": str(arr.dtype),
            }

        @classmethod
        def deserialize(cls, payload):
            arr = np.array(payload["data"], dtype=payload["dtype"])
            return arr

    NumpyArraySerializer.register(np.array, np.ndarray)



from __future__ import annotations

import typing

from fastapi import Request
from jinja2 import ChoiceLoader, FileSystemLoader
from jinja2_fragments.fastapi import Jinja2Blocks
from starlette.background import BackgroundTask

from xlwings_server.config import PACKAGE_DIR, PROJECT_DIR, settings

# Create Jinja2 loader that checks dist templates first (if exists),
# then project templates, then package templates
loaders = []
dist_templates = PROJECT_DIR / "dist" / "templates"
if dist_templates.exists() and settings.environment != "dev":
    loaders.append(FileSystemLoader(dist_templates))
loaders.extend(
    [
        FileSystemLoader(PROJECT_DIR / "templates"),
        FileSystemLoader(PACKAGE_DIR / "templates"),
    ]
)
loader = ChoiceLoader(loaders)


templates = Jinja2Blocks(
    directory=PROJECT_DIR / "templates",
    loader=loader,
    trim_blocks=True,
    lstrip_blocks=True,
)


# Align with FastAPIs changes to the function signature
def TemplateResponse(
    request: Request,
    name: str,
    context: dict[str, typing.Any] | None = None,
    status_code: int = 200,
    headers: typing.Mapping[str, str] | None = None,
    media_type: str | None = None,
    background: BackgroundTask | None = None,
    block_names: str | list[str] | None = None,
    **kwargs: typing.Any,
):
    if context is None:
        context = {}
    context["request"] = request
    context["settings"] = settings
    return templates.TemplateResponse(
        name,
        context,
        status_code,
        headers,
        media_type,
        background,
        block_name=block_names if isinstance(block_names, str) else None,
        block_names=block_names if isinstance(block_names, list) else [],
        **kwargs,
    )



import logging
import os
import re

from fastapi import APIRouter, Header, Request

from ..config import settings
from ..templates import TemplateResponse

router = APIRouter(prefix=settings.app_path)

logger = logging.getLogger(__name__)


def _get_manifest_response(
    request: Request, protocol: str, download: bool = False
) -> TemplateResponse:
    if settings.hostname:
        # Settings
        base_url = f"https://{settings.hostname}"
    elif os.getenv("RENDER_EXTERNAL_URL"):
        # Render.com
        base_url = os.getenv("RENDER_EXTERNAL_URL")
    elif os.getenv("WEBSITE_HOSTNAME"):
        # Azure Functions and Azure App Service
        base_url = f"https://{os.getenv('WEBSITE_HOSTNAME')}"
    elif os.getenv("CODESPACES"):
        # GitHub Codespaces
        base_url = f"https://{os.getenv('CODESPACE_NAME')}-8000.{os.getenv('GITHUB_CODESPACES_PORT_FORWARDING_DOMAIN')}"
    else:
        base_url = str(request.base_url)
        if protocol == "https":
            base_url = base_url.replace("http://", "https://")

    base_url = str(base_url).rstrip("/")

    manifest_ids = {
        "dev": settings.manifest_id_dev,
        "qa": settings.manifest_id_qa,
        "uat": settings.manifest_id_uat,
        "staging": settings.manifest_id_staging,
        "prod": settings.manifest_id_prod,
    }
    manifest_id = manifest_ids[settings.environment]

    headers = {}
    if download:
        sanitized_name = re.sub(
            r"[^a-z0-9]+", "-", settings.project_name.lower()
        ).strip("-")
        filename = f"{sanitized_name}-{settings.environment}.xml"
        headers["Content-Disposition"] = f"attachment; filename={filename}"

    return TemplateResponse(
        request=request,
        name="manifest.xml",
        context={
            "settings": settings,
            "base_url": base_url,
            "base_url_with_app_path": f"{base_url}{settings.app_path}",
            "manifest_id": manifest_id,
        },
        media_type="text/plain",
        headers=headers,
    )


@router.get("/manifest")
@router.get("/manifest.xml")
async def manifest(
    request: Request,
    protocol: str = Header(default="", alias="X-Forwarded-Proto"),
):
    return _get_manifest_response(request, protocol, download=False)


@router.get("/manifest/download")
async def manifest_download(
    request: Request,
    protocol: str = Header(default="", alias="X-Forwarded-Proto"),
):
    return _get_manifest_response(request, protocol, download=True)



from fastapi import APIRouter, Request

from ..config import settings
from ..templates import TemplateResponse

router = APIRouter(prefix=settings.app_path)


@router.get("/taskpane")
@router.get("/taskpane.html")
async def taskpane(request: Request):
    return TemplateResponse(request=request, name=settings.taskpane_html)



"""
Make sure to import the desired functions under __init__.py, e.g.:
from .example import *
"""

try:
    import matplotlib.pyplot as plt
except ImportError:
    plt = None
import sys
from pathlib import Path

import numpy as np
import xlwings as xw
from xlwings.server import script

from . import settings


@script
def hello_world(book: xw.Book):
    sheet = book.sheets.active
    cell = sheet["A1"]
    if cell.value == "Hello package!":
        cell.value = "Bye package!"
    else:
        cell.value = "Hello package!"


@script
def show_alert(book: xw.Book):
    # callback is optional and only required if you want a JS function to be called
    # when clicking a button. alertCallback is defined in app/static/js/core/examples.js
    book.app.alert(
        "This is an alert!",
        title="xlwings Server Alert",
        buttons="ok_cancel",
        callback="alertCallback",
    )


@script
def setup_custom_functions(book: xw.Book):
    prefix = f"{settings.functions_namespace}"
    if settings.environment != "prod":
        prefix += f"_{settings.environment}".upper()
    sheet = book.sheets.add()
    sheet["A3"].value = f'={prefix}.HELLO("xlwings")'
    sheet["A5"].value = f"={prefix}.STANDARD_NORMAL(3, 4)"
    sheet["A10"].value = f"={prefix}.CORREL(A5#)"
    if not settings.enable_wasm:
        sheet["A16"].value = f"={prefix}.TO_DF(A5#)"
        sheet["A18"].value = f"={prefix}.GET_HEALTHEXP()"
        sheet[
            "A20"
        ].value = f"""={prefix}.DF_QUERY(A18, "Country == 'Japan' and Year > 2017")"""
        sheet["A25"].value = f"={prefix}.VIEW(A18, 3)"
        sheet["A30"].value = f"={prefix}.STREAMING_RANDOM(3, 4)"
        sheet["A35"].value = f"={prefix}.GET_CURRENT_USER()"
        sheet[
            "A37"
        ].value = f'={prefix}.SQL("SELECT Year, Country FROM a WHERE Spending_USD < 4600", A20#)'
        sheet["A40"].value = f'={prefix}.HELLO_WITH_SCRIPT("xlwings")'
    sheet.activate()


@script
def show_plot(book: xw.Book):
    """Adopted from
    https://matplotlib.org/stable/plot_types/stats/hexbin.html#sphx-glr-plot-types-stats-hexbin-py
    """
    if not plt:
        raise xw.XlwingsError("You need to install Matplotlib for this example")
    plt.style.use("_mpl-gallery-nogrid")
    rng = np.random.default_rng()
    x = rng.standard_normal(5000)
    y = 1.2 * x + rng.standard_normal(5000) / 3
    fig, ax = plt.subplots()
    ax.hexbin(x, y, gridsize=20)
    ax.set(xlim=(-2, 2), ylim=(-3, 3))
    book.sheets.active.pictures.add(
        fig, anchor=book.sheets.active["A10"], update=True, name="mplot"
    )


@script
def show_error(book: xw.Book):
    raise xw.XlwingsError("This would be your error message")


# Unit tests
if settings.enable_tests:
    sys.path.append(str(Path(__file__).parent.parent.resolve()))
    from tests.e2e_custom_scripts import *



import json
import logging
import os
import sys
from functools import cache
from pathlib import Path

import socketio
from fastapi import FastAPI, status
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import PlainTextResponse
from fastapi.staticfiles import StaticFiles
from xlwings import XlwingsError

# CRITICAL: Setup sys.path for user overrides BEFORE importing user modules
# This is done here (not in CLI) so it works when uvicorn imports this module
if project_dir := os.getenv("XLWINGS_PROJECT_DIR"):
    project_dir = Path(project_dir)
    if str(project_dir) not in sys.path:
        sys.path.insert(0, str(project_dir))

from xlwings_server.config import PACKAGE_DIR, PROJECT_DIR, settings
from xlwings_server.object_handles import ObjectCacheConverter
from xlwings_server.routers import socketio as socketio_router
from xlwings_server.routers.manifest import router as manifest_router
from xlwings_server.routers.root import router as root_router
from xlwings_server.routers.taskpane import router as taskpane_router
from xlwings_server.routers.xlwings import router as xlwings_router
from xlwings_server.templates import templates

# Logging
logging.basicConfig(level=settings.log_level.upper())
logger = logging.getLogger(__name__)

logger.info(f"Running in '{'Wasm' if settings.enable_wasm else 'Server'}' mode.")

# App
app = FastAPI(docs_url=None, redoc_url=None, openapi_url=None)


# Starlette's url_for returns fully qualified URLs causing issues if the reverse proxy
# handles TLS and the app runs on http (https://github.com/encode/starlette/issues/843)
def url_for(name: str, **path_params):
    """Wrapper around url_path_for that strips leading slashes from path parameters"""
    if name == "static" and "path" in path_params:
        # Strip leading slash from static file paths to avoid double slashes
        path_params["path"] = path_params["path"].lstrip("/")
    return app.url_path_for(name, **path_params)


templates.env.globals["url_for"] = url_for

# Register Converter
ObjectCacheConverter.register(object, "object", "obj")


# Custom StaticFiles with file-level override support
class OverridableStaticFiles(StaticFiles):
    """StaticFiles that checks user directory first, then falls back to package directory"""

    def __init__(self, package_directory: Path, user_directory: Path, **kwargs):
        self.package_dir = package_directory
        self.user_dir = user_directory
        super().__init__(directory=str(package_directory), **kwargs)

    def lookup_path(self, path: str) -> tuple[str, os.stat_result | None]:
        # Check user directory first
        user_path = self.user_dir / path
        if user_path.is_file():
            return str(user_path), os.stat(user_path)
        # Fallback to package directory
        return super().lookup_path(path)


# CORS: Office Scripts and custom functions in Excel on the web require CORS
# Using app.add_middleware won't add the CORS headers if you handle the root "Exception"
# in an exception handler (it would require a specific exception type).
cors_app = CORSMiddleware(
    app=app,
    allow_origins=settings.cors_allow_origins,
    allow_methods=["POST"],
    allow_headers=["*"],
    allow_credentials=False,
)
main_app = cors_app

# Socket.io
if settings.enable_socketio:
    sio_app = socketio.ASGIApp(
        socketio_router.sio,
        # Only forward ASGI traffic if there's no message queue and hence 1 worker setup
        cors_app if not settings.socketio_message_queue_url else None,
        socketio_path=f"{settings.app_path}/socket.io",
    )
    main_app = sio_app if not settings.socketio_message_queue_url else cors_app


# Routers
app.include_router(root_router)
app.include_router(xlwings_router)
app.include_router(taskpane_router)
app.include_router(manifest_router)


# User routers (optional custom router)
try:
    from routers import custom

    if hasattr(custom, "router"):
        app.include_router(custom.router)
        logger.info("Registered custom user router")
except ModuleNotFoundError:
    logger.debug("No custom router found (routers/custom.py)")
except Exception:
    logger.exception("Failed to load custom router from routers/custom.py")


# Security headers
@cache
def read_security_headers():
    with open(settings.base_dir / "security_headers.json", "r") as f:
        data = json.load(f)
    return data


@app.middleware("http")
async def add_security_headers(request, call_next):
    # https://owasp.org/www-project-secure-headers/index.html#configuration-proposal
    # https://owasp.org/www-project-secure-headers/ci/headers_add.json
    response = await call_next(request)
    if not settings.add_security_headers and settings.environment == "dev":
        # Prevent caching in dev even if security headers are switched off
        response.headers["Cache-Control"] = "no-store, max-age=0"
    if settings.add_security_headers:
        data = read_security_headers()

        # Extract file extension from request URL
        file_ext = request.url.path.split(".")[-1].lower()
        image_ext = ("jpg", "jpeg", "png", "gif", "bmp", "webp", "svg")

        for header in data["headers"]:
            # Excel on Windows doesn't display the icons in the ribbon otherwise
            if header["name"] == "Cache-Control" and file_ext in image_ext:
                continue
            if header["name"] not in (
                "Permissions-Policy",  # experimental
                "Clear-Site-Data",  # too aggressive
                "Content-Security-Policy",  # provide via XLWINGS_CUSTOM_HEADERS
            ):
                # Permissions-Policy headers are experimental
                # Clear-Site-Data is too aggressive
                response.headers[header["name"]] = header["value"]

        if settings.enable_excel_online:
            response.headers["Cross-Origin-Resource-Policy"] = "cross-origin"
            del response.headers["X-Frame-Options"]
        if settings.cdn_officejs:
            response.headers["Cross-Origin-Resource-Policy"] = "cross-origin"
            del response.headers["Cross-Origin-Embedder-Policy"]

    # Custom headers
    for header_name, header_value in settings.custom_headers.items():
        response.headers[header_name] = header_value

    return response


# Static files: in prod might be served by something like nginx or via
# https://github.com/matthiask/blacknoise or https://github.com/Archmonger/ServeStatic
# Auto-detect dist/ directory for production builds with hashed filenames
dist_static = PROJECT_DIR / "dist" / "static"
if dist_static.exists() and settings.environment != "dev":
    app.mount(
        settings.static_url_path,
        StaticFiles(directory=dist_static),
        name="static",
    )
    logger.info("Serving static files from dist/ (production build)")
else:
    app.mount(
        settings.static_url_path,
        OverridableStaticFiles(
            package_directory=PACKAGE_DIR / "static",
            user_directory=PROJECT_DIR / "static",
        ),
        name="static",
    )


if settings.enable_wasm:
    # For xlwings Wasm development
    app.mount(
        # Use the same path prefix as for static files
        settings.static_url_path.replace("static", "wasm"),
        OverridableStaticFiles(
            package_directory=PACKAGE_DIR / "wasm",
            user_directory=PROJECT_DIR / "wasm",
        ),
        name="wasm",
    )

    app.mount(
        # Use the same path prefix as for static files
        settings.static_url_path.replace("static", "custom_functions"),
        OverridableStaticFiles(
            package_directory=PACKAGE_DIR / "custom_functions",
            user_directory=PROJECT_DIR / "custom_functions",
        ),
        name="custom_functions",
    )

    app.mount(
        # Use the same path prefix as for static files
        settings.static_url_path.replace("static", "custom_scripts"),
        OverridableStaticFiles(
            package_directory=PACKAGE_DIR / "custom_scripts",
            user_directory=PROJECT_DIR / "custom_scripts",
        ),
        name="custom_scripts",
    )

if settings.environment == "dev":
    # Don't cache static files
    StaticFiles.is_not_modified = lambda *args, **kwargs: False


# Exception handlers
@app.exception_handler(XlwingsError)
async def xlwings_exception_handler(request, exception):
    logger.error(exception)
    msg = str(exception)
    return PlainTextResponse(msg, status_code=status.HTTP_500_INTERNAL_SERVER_ERROR)


@app.exception_handler(Exception)
async def exception_handler(request, exception):
    logger.error(exception)
    if settings.environment == "dev":
        msg = repr(exception)
    else:
        msg = "An error occurred."
    return PlainTextResponse(msg, status_code=status.HTTP_500_INTERNAL_SERVER_ERROR)



from config import settings  # noqa: F401, I001 Must be first import to load env vars
import platform
import traceback
import importlib.util
import sys
import contextlib
from io import StringIO
import ast
import inspect
from textwrap import dedent

import custom_functions
import custom_scripts
import js  # type: ignore
import pyodide_js  # type: ignore
import xlwings as xw
import xlwings.server
from pyodide.ffi import to_js  # type: ignore

try:
    import matplotlib as mpl

    mpl.use("agg")
except ImportError:
    mpl = None


# xlwings <0.33.13 doesn't have custom_scripts_meta
if not hasattr(xlwings.server, "custom_scripts_meta"):
    xlwings.server.custom_scripts_meta = lambda module: []


class HtmlOutput:
    def __init__(self, div_id):
        self.div_id = div_id
        self.buffer = StringIO()

    def write(self, text):
        self.buffer.write(text)
        content = self.buffer.getvalue()
        div = js.document.getElementById(self.div_id)
        if div:
            js.document.getElementById(self.div_id).innerHTML = f"<pre>{content}</pre>"
            # Scroll to bottom
            outputContainer = js.document.querySelector("#output-container")
            if outputContainer:
                outputContainer.scrollTop = outputContainer.scrollHeight

    def flush(self):
        pass


if settings.is_official_lite_addin:
    html_output = HtmlOutput("output")
    try:
        import pyodide_http  # type: ignore
    except ImportError:
        pyodide_http = None
    if pyodide_http:
        pyodide_http.patch_all()  # Fixes pd.read_csv()


# Print Python and Pyodide versions
def print_versions():
    if settings.is_official_lite_addin:
        html_output = HtmlOutput("output")
        with (
            contextlib.redirect_stdout(html_output),
            contextlib.redirect_stderr(html_output),
        ):
            print(
                f"Python {platform.python_version()}|Pyodide {pyodide_js.version}|xlwings {xw.__version__}"
            )


def create_module_from_string(module_string, module_name, html_output=None):
    spec = importlib.util.spec_from_loader(
        module_name,
        loader=None,
    )
    module = importlib.util.module_from_spec(spec)
    if html_output:
        with (
            contextlib.redirect_stdout(html_output),
            contextlib.redirect_stderr(html_output),
        ):
            exec(module_string, module.__dict__)
    else:
        exec(module_string, module.__dict__)
    sys.modules[module_name] = module
    return module


async def custom_functions_call(data):
    if settings.is_official_lite_addin:
        import main_editor

        module = main_editor
    else:
        module = custom_functions

    data = data.to_py()
    try:
        if settings.is_official_lite_addin:
            with (
                contextlib.redirect_stdout(html_output),
                contextlib.redirect_stderr(html_output),
            ):
                result = await xlwings.server.custom_functions_call(
                    data,
                    module=module,
                )
        else:
            result = await xlwings.server.custom_functions_call(
                data,
                module=module,
            )
    except Exception as e:
        result = {"error": str(e), "details": traceback.format_exc()}
    # Note: converts None to undefined
    return to_js(result, dict_converter=js.Object.fromEntries)


async def custom_scripts_call(data, script_name):
    if settings.is_official_lite_addin:
        import main_editor

        module = main_editor
    else:
        module = custom_scripts

    book = xw.Book(json=data.to_py())
    try:
        if settings.is_official_lite_addin:
            with (
                contextlib.redirect_stdout(html_output),
                contextlib.redirect_stderr(html_output),
            ):
                book = await xlwings.server.custom_scripts_call(
                    module=module,
                    script_name=script_name,
                    typehint_to_value={xw.Book: book},
                )
        else:
            book = await xlwings.server.custom_scripts_call(
                module=module,
                script_name=script_name,
                typehint_to_value={xw.Book: book},
            )
        result = book.json()
        return to_js(result, dict_converter=js.Object.fromEntries)
    except Exception as e:
        error_traceback = traceback.format_exc()
        if settings.is_official_lite_addin:
            error_output = HtmlOutput(
                "output"
            )  # Clear output before writing out an error
            error_output.write(f'<span style="color: red">{error_traceback}</span>')
            return to_js({"actions": []}, dict_converter=js.Object.fromEntries)
        else:
            result = {"error": str(e), "details": error_traceback}
            return to_js(result, dict_converter=js.Object.fromEntries)


def custom_scripts_meta():
    if settings.is_official_lite_addin:
        import main_editor

        module = main_editor
    else:
        module = custom_scripts
    scripts_meta = xlwings.server.custom_scripts_meta(module)
    return to_js(scripts_meta, dict_converter=js.Object.fromEntries)


def get_xlwings_scripts(code_string, script_button_text):
    try:
        tree = ast.parse(code_string)
    except:  # noqa: E722
        return to_js([script_button_text])
    xlwings_functions = []

    for node in ast.walk(tree):
        if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
            for decorator in node.decorator_list:
                decorator_str = ast.unparse(decorator)
                if "script" in decorator_str:
                    xlwings_functions.append(node.name)
                    break

    # If currently selected script is still here, return it at beginning
    if script_button_text in xlwings_functions:
        remaining = [x for x in xlwings_functions if x != script_button_text]
        return to_js([script_button_text] + remaining)

    return to_js(xlwings_functions)


html_output = HtmlOutput("output")


def custom_functions_meta(module_string):
    try:
        module = create_module_from_string(module_string, "main")
    except Exception as e:  # noqa: E722
        js.console.log(f"Parsing error [custom_functions_meta]: {e}")
        return to_js({}, dict_converter=js.Object.fromEntries)
    return to_js(
        xlwings.server.custom_functions_meta(module),
        dict_converter=js.Object.fromEntries,
    )


def custom_functions_code(module_string):
    try:
        module = create_module_from_string(module_string, "main")
    except Exception as e:  # noqa: E722
        js.console.log(f"Parsing error [custom_functions_code]: {e}")
        return ""
    code = ""
    for _, obj in inspect.getmembers(module):
        if hasattr(obj, "__xlfunc__"):
            xlfunc = obj.__xlfunc__
            func_name = xlfunc["name"]
            streaming = "true" if inspect.isasyncgenfunction(obj) else "false"
            code += dedent(
                f"""\
            async function {func_name}() {{
                let args = ["{func_name}", {streaming}]
                args.push.apply(args, arguments);
                return await base.apply(null, args);
            }}
            CustomFunctions.associate("{func_name.upper()}", {func_name});
            """
            )
    return code



from .framework import Serializer, custom_decoder


class DictionarySerializer(Serializer):
    name = "custom_dict_serializer"

    @classmethod
    def serialize(cls, obj):
        # Convert dictionary with non-string keys (e.g., datetime) to a list of
        # [key, value] pairs
        items = [[k, v] for k, v in obj.items()]
        return {
            "data": items,
            "serializer": cls.name,
        }

    @classmethod
    def deserialize(cls, payload):
        # Convert back from list of pairs to dictionary
        items = payload["data"]
        return {custom_decoder(k): custom_decoder(v) for k, v in items}


# Register the serializer for dict type
DictionarySerializer.register(dict)



import redis  # TODO: use redis.asyncio when converters can be used async

from .config import settings

# Redis
redis_pool = None
if settings.object_cache_url:
    redis_pool = redis.ConnectionPool.from_url(settings.object_cache_url)


def get_redis_client():
    if redis_pool is None:
        yield None
    else:
        client = redis.Redis.from_pool(redis_pool)
        try:
            yield client
        finally:
            client.close()



# Try to import User model from project directory first (user override)
# Fall back to package location (default implementation)
try:
    from models.user import User
except ModuleNotFoundError:
    from .user import User


class CurrentUser(User):
    """
    Will deliver the current user when used as type hint in a custom function/script
    """

    pass



import logging
from typing import Any

from pydantic import BaseModel, Field, model_validator

try:
    from typing import Self
except ImportError:
    # Python < 3.11
    from typing_extensions import Self

logger = logging.getLogger(__name__)


class User(BaseModel):
    """
    User model. This is set up so that it populates most fields from claims
    as returned by OIDC tokens (e.g., Entra ID). You can also use the model by setting
    the attributes directly though.
    """

    id: str | None = None
    name: str | None = None
    domain: str | None = None
    email: str | None = None
    roles: list[str] = []
    ip_address: str | None = None
    claims: dict[str, Any] = Field({}, repr=False)

    @model_validator(mode="after")
    def populate_from_claims(self) -> Self:
        if self.claims:
            self.id = self.id or self.claims.get("oid")
            self.name = self.name or self.claims.get("name")
            self.email = self.email or self.claims.get("preferred_username")
            self.roles = self.roles or self.claims.get("roles", [])
            if not self.domain and self.email and "@" in self.email:
                self.domain = self.email.split("@")[1]
        return self

    async def has_required_roles(self, required_roles: list[str] | None = None):
        if required_roles:
            if set(required_roles).issubset(self.roles):
                logger.info(f"User has required roles: {self.name}")
                return True
            else:
                return False
        else:
            return True

    async def is_authorized(self):
        """Method that can be overridden to implement a global authorization logic"""
        return True



import datetime as dt

# Registry, holding type or serializer_name vs. serializer combinations
serializers = {}


# Base class
class Serializer:
    @classmethod
    def serialize(cls, obj):
        raise NotImplementedError()

    @classmethod
    def deserialize(cls, payload):
        raise NotImplementedError()

    @classmethod
    def register(cls, *types):
        serializers[cls.name] = cls
        for type in types:
            serializers[type] = cls


# Custom encoders/decoders
def custom_encoder(obj):
    if isinstance(obj, dt.datetime):
        return obj.isoformat()
    valid_types = tuple(cls for cls in serializers.keys() if isinstance(cls, type))
    if isinstance(obj, valid_types):
        serializer = serializers.get(type(obj))
        if serializer:
            return serializer.serialize(obj)
    raise TypeError(f"Object of type {type(obj)} is not JSON serializable")


def custom_decoder(obj):
    if isinstance(obj, list):
        return [custom_decoder(item) for item in obj]
    elif isinstance(obj, dict):
        serializer = serializers.get(obj.get("serializer"))
        if serializer:
            return serializer.deserialize(obj)
        return {key: custom_decoder(value) for key, value in obj.items()}
    elif isinstance(obj, str):
        try:
            return dt.datetime.fromisoformat(obj)
        except ValueError:
            return obj
    else:
        return obj



"""Build-time utilities for xlwings Server.

This package contains tools used during build and deployment processes,
such as static file hashing for cache-busting.
"""

from .static_file_hasher import StaticFileHasher

__all__ = ["StaticFileHasher"]
